# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
"""
Provides a modal dialog box for logging in to AWS Deadline Cloud.

Example code:
    from deadline.client.ui.dialogs import DeadlineLoginDialog
    if DeadlineLoginDialog.login(parent=self):
        print("Logged in successfully.")
    else:
        print("Failed to log in.")
"""

__all__ = ["DeadlineLoginDialog"]

import html
import threading
from configparser import ConfigParser
from typing import Optional

from qtpy.QtCore import Signal
from .._utils import tr
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QMessageBox,
    QWidget,
)

from ... import api
from ...api._session import AwsCredentialsSource


class DeadlineLoginDialog(QMessageBox):
    """
    A modal dialog box for logging in to AWS Deadline Cloud. The return value
    of the static DeadlineLoginDialog.login() and the modal exec()
    is True when the login is successful, False otherwise.

    Example code:
        if DeadlineLoginDialog.login(parent=self):
            print("Logged in successfully.")
        else:
            print("Failed to log in.")
    """

    # This signal is sent when the background login thread raises an exception.
    login_thread_exception = Signal(BaseException)
    # This signal is sent when the background login thread wants to change the
    # displayed message.
    login_thread_message = Signal(str)
    # This signal is sent when the background login thread succeeds.
    login_thread_succeeded = Signal(str)

    @staticmethod
    def login(
        parent: Optional[QWidget] = None,
        force_refresh=False,
        close_on_success=True,
        config: Optional[ConfigParser] = None,
    ) -> bool:
        """
        Static method that runs the Deadline Login Dialog. Returns True for
        a successful login, False otherwise.

        Args:
            force_refresh (bool, default False): Forces a re-login even when already authorized.
            close_on_success (bool, default True): Closes the dialog on successful login, instead
                   of showing a "successfully logged in" message.
            config (ConfigParser, optional): The AWS Deadline Cloud configuration
                    object to use instead of the config file.
        """
        deadline_login = DeadlineLoginDialog(
            parent=parent,
            close_on_success=close_on_success,
            config=config,
        )
        return deadline_login.exec_()

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        close_on_success=True,
        config: Optional[ConfigParser] = None,
    ) -> None:
        super().__init__(parent=parent)

        self.close_on_success = close_on_success
        self.config = config
        self.canceled = False
        self.__login_thread: Optional[threading.Thread] = None
        self.login_thread_exception.connect(self.handle_login_thread_exception)
        self.login_thread_message.connect(self.handle_login_thread_message)
        self.login_thread_succeeded.connect(self.handle_login_thread_succeeded)
        self.buttonClicked.connect(self.on_button_clicked)

        self.setWindowTitle(tr("Log in to AWS Deadline Cloud"))
        self.setText(tr("Logging you in..."))
        self.setStandardButtons(QMessageBox.Cancel)

        self._start_login()

    def _login_background_thread(self):
        """
        This function gets started in a background thread to run the login handshake. It
        polls the `self.canceled` flag for cancellation, and fills in details about the
        login when they become available.
        """
        try:

            def on_pending_authorization(**kwargs):
                if (
                    kwargs["credentials_source"]
                    == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN
                ):
                    self.login_thread_message.emit(
                        tr("Opening Deadline Cloud monitor. Please log in before returning here.")
                    )

            def on_cancellation_check():
                return self.canceled

            success_message = api.login(
                on_pending_authorization,
                on_cancellation_check,
                config=self.config,
            )
            self.login_thread_succeeded.emit(success_message)
        except BaseException as e:
            # Send the exception to the dialog
            self.login_thread_exception.emit(e)

    def _start_login(self) -> None:
        """
        Starts the background login thread.
        """

        self.__login_thread = threading.Thread(
            target=self._login_background_thread, name="AWS Deadline Cloud login thread"
        )
        self.__login_thread.start()

    def handle_login_thread_exception(self, e: BaseException) -> None:
        """
        Handles the signal sent from the background login thread when
        an exception is thrown.
        """
        self.setStandardButtons(QMessageBox.Close)
        self.setIcon(QMessageBox.Warning)
        self.setText(
            tr("Failed to log in to AWS Deadline Cloud:<br/><br/>{error}").format(
                error=html.escape(str(e))
            )
        )

    def handle_login_thread_message(self, message: str) -> None:
        """
        Handles the signal sent from the background login thread when
        the message needs to be set.
        """
        self.setText(message)

    def handle_login_thread_succeeded(self, success_message: str) -> None:
        """
        Handles the signal sent from the background login thread when
        the login has succeeded.
        """
        if self.close_on_success:
            # Effectively clicks on "OK"
            self.accept()
        else:
            self.setStandardButtons(QMessageBox.Ok)
            self.setIcon(QMessageBox.Information)
            self.setText(
                tr("Successfully logged into: <br/><br/>{profile}").format(
                    profile=html.escape(success_message)
                )
            )

    def on_button_clicked(self, button):
        if self.standardButton(button) == QMessageBox.Cancel:
            # Tell the login thread to cancel, then wait for it.
            self.canceled = True
            if self.__login_thread:
                while self.__login_thread.is_alive():
                    QApplication.instance().processEvents()  # type: ignore[union-attr]

    def exec_(self) -> bool:
        """
        Runs the modal login dialog, returning True if the login was
        successful, False otherwise.
        """
        return super().exec_() == QMessageBox.Ok



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "DeadlineConfigDialog",
    "DeadlineLoginDialog",
    "SubmitJobProgressDialog",
    "SubmitJobToDeadlineDialog",
    "JobBundlePurpose",
]

from ._types import JobBundlePurpose
from .deadline_config_dialog import DeadlineConfigDialog
from .deadline_login_dialog import DeadlineLoginDialog
from .submit_job_progress_dialog import SubmitJobProgressDialog
from .submit_job_to_deadline_dialog import SubmitJobToDeadlineDialog



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
The AWS Deadline Cloud CLI interface.
"""

import logging
import os
import sys
from logging import getLogger
from typing import Optional

import click

from .. import version
from ..api._session import session_context
from ..config import get_setting, get_setting_default
from ._common import _PROMPT_WHEN_COMPLETE

logger = getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
_DEADLINE_LOG_LEVELS = [
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
]  # Log Levels AWS Deadline Cloud Allows


def _get_default_log_level() -> str:
    """
    Get the default log level from the config file.
    """
    # Set the default log level based on the setting, must do here so we can pass into the click option
    _SETTING_LOG_LEVEL = get_setting("settings.log_level").upper()
    _DEFAULT_LOG_LEVEL = get_setting_default("settings.log_level")
    _CLI_DEFAULT_LOG_LEVEL = _DEFAULT_LOG_LEVEL
    if _SETTING_LOG_LEVEL not in _DEADLINE_LOG_LEVELS:
        logger.warning(
            f"Log Level '{_SETTING_LOG_LEVEL}' not in {_DEADLINE_LOG_LEVELS}. Defaulting to {_DEFAULT_LOG_LEVEL}"
        )
    else:
        _CLI_DEFAULT_LOG_LEVEL = _SETTING_LOG_LEVEL
    return _CLI_DEFAULT_LOG_LEVEL


class ContextTrackingCommand(click.Command):
    """
    Adds the current CLI command name to User Agent headers in boto requests
    """

    def invoke(self, ctx: click.Context):
        # This is a global variable used to modify User Agent header in the default boto config
        session_context["cli-command-name"] = ctx.command_path.replace(" ", ".")
        return super().invoke(ctx)


class ContextTrackingGroup(click.Group):
    """
    Adds the current CLI command name to User Agent headers in boto requests
    """

    # Special value documented in Click to make this group class the default
    # See https://click.palletsprojects.com/en/stable/api/#click.Group.group_class
    group_class = type

    # Special value documented in Click to make this command class the default
    # See https://click.palletsprojects.com/en/stable/api/#click.Group.command_class
    command_class = ContextTrackingCommand


@click.group(cls=ContextTrackingGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version, prog_name="deadline")
@click.option(
    "--log-level",
    type=click.Choice(_DEADLINE_LOG_LEVELS, case_sensitive=False),
    default=None,
    help="Set the logging level.",
)
@click.option(
    "--redirect-output",
    help="Redirects stdout and stderr messages to append to the specified file. "
    "Useful for the 'deadlinew' command which does not produce terminal output by default on Windows.",
)
@click.option(
    "--redirect-mode",
    type=click.Choice(["append", "replace"], case_sensitive=False),
    default="append",
    help="When using the --redirect-output option, controls whether to append to or replace the output file.",
)
@click.pass_context
def deadline(
    ctx: click.Context, log_level: Optional[str], redirect_output: str, redirect_mode: str
):
    """
    The `deadline` command provides functionality to interact with [AWS Deadline Cloud].

    For example, you can submit jobs to a queue with `deadline bundle submit` or
    `deadline bundle gui-submit`, monitor the status of a job with `deadline job get` and
    `deadline job logs`, wait for job completion with `deadline job wait`, then retrieve
    the output with `deadline job download-output`. You can also use
    `deadline queue sync-output` as an alternative to downloading individual jobs,
    to retrieve all the output of jobs in a queue over time.

    The command works with any local AWS credentials you have configured, or together with
    [Deadline Cloud monitor] to use AWS credentials from logging into the identity provider
    configured for your farm.

    [AWS Deadline Cloud]: https://aws.amazon.com/deadline-cloud/
    [Deadline Cloud monitor]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/working-with-deadline-monitor.html
    """
    if redirect_output:
        # Set both stdout and stderr to write to the specified file, writing in line buffering mode
        if redirect_mode == "append":
            open_mode = "a"
        else:
            open_mode = "w"
        sys.stdout = sys.stderr = open(redirect_output, open_mode, encoding="utf-8", buffering=1)
    elif sys.stdout and os.name == "nt":
        # Force the output encoding to be UTF-8 on Windows. Commands like `deadline job logs` print out logs that
        # can include unicode, and this enables output without errors.
        sys.stdout.reconfigure(encoding="utf-8")  # type: ignore
    if log_level is None:
        log_level = _get_default_log_level()

    logging.basicConfig(level=log_level)
    if log_level == "DEBUG":
        logger.debug("Debug logging is on")

    ctx.ensure_object(dict)
    # By default don't prompt when the operation is complete
    ctx.obj[_PROMPT_WHEN_COMPLETE] = False



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from typing import Any, Callable, List, Optional

from deadline.client.api._job_attachment import _hash_attachments
from deadline.job_attachments.asset_manifests.base_manifest import BaseAssetManifest
from deadline.job_attachments.upload import S3AssetManager


def _create_manifest_for_single_root(
    files: List[str],
    root: str,
    print_function_callback: Callable[[Any], None] = lambda msg: None,
) -> Optional[BaseAssetManifest]:
    """
    Shared logic to create a manifest file from a single root.
    :param files: Input files to create a manifest with.
    :param root: Asset root of the files.
    :param logger: Click logger for stdout.
    :return
    """
    # Placeholder Asset Manager
    asset_manager = S3AssetManager()

    from deadline.client.cli._common import _ProgressBarCallbackManager

    hash_callback_manager = _ProgressBarCallbackManager(length=100, label="Hashing Attachments")

    upload_group = asset_manager.prepare_paths_for_upload(
        input_paths=files, output_paths=[root], referenced_paths=[]
    )
    # We only provided 1 root path, so output should only have 1 group.
    assert len(upload_group.asset_groups) == 1

    if upload_group.asset_groups:
        _, manifests = _hash_attachments(
            asset_manager=asset_manager,
            asset_groups=upload_group.asset_groups,
            total_input_files=upload_group.total_input_files,
            total_input_bytes=upload_group.total_input_bytes,
            print_function_callback=print_function_callback,
            hashing_progress_callback=hash_callback_manager.callback,
        )

    if not manifests or len(manifests) == 0:
        print_function_callback("No manifest generated")
        return None
    else:
        # This is a hard failure, we are snapshotting 1 directory.
        assert len(manifests) == 1

        # Return the generated manifest.
        return manifests[0].asset_manifest



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline job` commands.
"""

from __future__ import annotations
import json
import logging
from configparser import ConfigParser
from pathlib import Path
import os
import re
import sys
from typing import Callable, Optional, Union
import datetime
from typing import Any
import textwrap

import click
from botocore.exceptions import ClientError

from ...api._session import _modified_logging_level
from ....job_attachments.download import OutputDownloader
from ....job_attachments.models import (
    FileConflictResolution,
    JobAttachmentS3Settings,
    PathFormat,
)
from ....job_attachments._utils import (
    WINDOWS_MAX_PATH_LENGTH,
    _is_windows_long_path_registry_enabled,
)
from ....job_attachments.progress_tracker import (
    DownloadSummaryStatistics,
    ProgressReportMetadata,
)
from ....job_attachments._path_summarization import (
    human_readable_file_size,
    summarize_path_list,
)

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError, DeadlineOperationTimedOut
from .._common import (
    _apply_cli_options_to_config,
    _cli_object_repr,
    _handle_error,
    _suggest_resources_on_client_error,
)
from .._main import deadline as main
from ._sigint_handler import SigIntHandler
from ...api._session import get_default_client_config
from .._timestamp_formatter import TimestampFormat, TimestampFormatter
from ._job_helpers import _resolve_job_search, _print_job_details

logger = logging.getLogger("deadline.client.cli")

JSON_MSG_TYPE_TITLE = "title"
JSON_MSG_TYPE_PRESUMMARY = "presummary"
JSON_MSG_TYPE_PATH = "path"
JSON_MSG_TYPE_PATHCONFIRM = "pathconfirm"
JSON_MSG_TYPE_PROGRESS = "progress"
JSON_MSG_TYPE_SUMMARY = "summary"
JSON_MSG_TYPE_ERROR = "error"
JSON_MSG_TYPE_WARNING = "warning"


def _parse_session_action_id(session_action_id: str) -> str:
    """
    Parse a session action ID and extract the session ID.

    Session action ID format: sessionaction-{uuid}-{number}
    Session ID format: session-{uuid}

    Args:
        session_action_id: The session action ID to parse

    Returns:
        The derived session ID

    Raises:
        DeadlineOperationError: If the session action ID format is invalid
    """
    pattern = r"^sessionaction-([0-9a-f]{32})-\d+$"
    match = re.match(pattern, session_action_id)

    if not match:
        raise DeadlineOperationError(
            f"Invalid session action ID format: '{session_action_id}'. "
            f"Expected format: sessionaction-{{uuid}}-{{number}}"
        )

    uuid = match.group(1)
    session_id = f"session-{uuid}"

    return session_id


# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


@main.group(name="job")
@_handle_error
def cli_job():
    """
    Commands to work with [Deadline Cloud jobs] in a [queue].

    Use the `deadline bundle submit` or `deadline bundle gui-submit` commands to create a job
    from a job bundle.

    [Deadline Cloud jobs]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    [queue]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/queues.html
    """


@cli_job.command(name="list")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--page-size", default=5, help="The number of jobs to load at a time.")
@click.option("--item-offset", default=0, help="The index of the job to start listing from.")
@_handle_error
def job_list(page_size, item_offset, **args):
    """
    Lists the [Deadline Cloud jobs] in the queue.

    [Deadline Cloud jobs]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    try:
        response = deadline.search_jobs(
            farmId=farm_id,
            queueIds=[queue_id],
            itemOffset=item_offset,
            pageSize=page_size,
            sortExpressions=[{"fieldSort": {"name": "CREATED_AT", "sortOrder": "DESCENDING"}}],
        )
    except ClientError as exc:
        suggestion = _suggest_resources_on_client_error(
            exc, farm_id=farm_id, queue_id=queue_id, config=config
        )
        raise DeadlineOperationError(
            f"Failed to get Jobs from Deadline:\n{exc}{suggestion}"
        ) from exc

    total_results = response["totalResults"]

    # Select which fields to print and in which order
    name_field = "displayName"
    if len(response["jobs"]) and "name" in response["jobs"][0]:
        name_field = "name"
    structured_job_list = [
        {
            field: job.get(field, "")
            for field in [
                name_field,
                "jobId",
                "taskRunStatus",
                "startedAt",
                "endedAt",
                "createdBy",
                "createdAt",
            ]
        }
        for job in response["jobs"]
    ]

    click.echo(
        f"Displaying {len(structured_job_list)} of {total_results} Jobs starting at {item_offset}"
    )
    click.echo()
    click.echo(_cli_object_repr(structured_job_list))


@cli_job.command(name="get")
@click.argument("search_term", required=False)
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to get.")
@_handle_error
def job_get(search_term: Optional[str], **args):
    """
    Get the details of a [Deadline Cloud job], or search for jobs with a search term.

    SEARCH_TERM can be a job ID (job-xxx) or a search string to find matching jobs.
    If exactly one job matches, shows full details. If multiple match, shows a summary list.
    If no arguments provided, shows the default job from config.

    [Deadline Cloud job]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    """
    # Check if search_term is actually a job ID
    if search_term and re.match(r"^job-[0-9a-f]{32}$", search_term):
        args["job_id"] = search_term
        search_term = None

    if search_term:
        # Search with term - don't require job_id
        config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)
        found_job_id = _resolve_job_search(config, search_term)
        if found_job_id:
            _print_job_details(config, found_job_id)
    else:
        # Get job by ID (from arg or config default)
        config = _apply_cli_options_to_config(
            required_options={"farm_id", "queue_id", "job_id"}, **args
        )
        job_id = config_file.get_setting("defaults.job_id", config=config)
        if job_id is None:
            raise DeadlineOperationError("Missing job ID. Provide a job ID or search term.")
        _print_job_details(config, job_id)


@cli_job.command(name="cancel")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to cancel.")
@click.option(
    "--mark-as",
    type=click.Choice(["SUSPENDED", "CANCELED", "FAILED", "SUCCEEDED"], case_sensitive=False),
    default="CANCELED",
    help="The status to apply to all active tasks in the job.",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@_handle_error
def job_cancel(mark_as: str, yes: bool, **args):
    """
    Cancel a [Deadline Cloud job] from running, optionally marking it with an alternative status such
    as SUSPENDED, SUCCEEDED or FAILED.

    [Deadline Cloud job]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    mark_as = mark_as.upper()

    deadline = api.get_boto3_client("deadline", config=config)

    # Print a summary of the job to cancel
    try:
        job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    except ClientError as exc:
        suggestion = _suggest_resources_on_client_error(
            exc, farm_id=farm_id, queue_id=queue_id, config=config
        )
        raise DeadlineOperationError(
            f"Failed to get Job from Deadline:\n{exc}{suggestion}"
        ) from exc
    # Remove the zero-count status counts
    job["taskRunStatusCounts"] = {
        name: count for name, count in job["taskRunStatusCounts"].items() if count != 0
    }
    # Filter the fields to a summary
    filtered_job = {
        field: job.get(field, "")
        for field in [
            "name",
            "jobId",
            "taskRunStatus",
            "taskRunStatusCounts",
            "startedAt",
            "endedAt",
            "createdBy",
            "createdAt",
        ]
    }
    click.echo(_cli_object_repr(filtered_job))

    # Ask for confirmation about canceling this job.
    if not (
        yes or config_file.str2bool(config_file.get_setting("settings.auto_accept", config=config))
    ):
        if mark_as == "CANCELED":
            cancel_message = "Are you sure you want to cancel this job?"
        else:
            cancel_message = (
                f"Are you sure you want to cancel this job and mark its taskRunStatus as {mark_as}?"
            )
        # We explicitly require a yes/no response, as this is an operation that will interrupt the work in progress
        # on their job.
        if not click.confirm(
            cancel_message,
            default=None,
        ):
            click.echo("Job not canceled.")
            sys.exit(1)

    if mark_as == "CANCELED":
        click.echo("Canceling job...")
    else:
        click.echo(f"Canceling job and marking as {mark_as}...")
    deadline.update_job(farmId=farm_id, queueId=queue_id, jobId=job_id, targetTaskRunStatus=mark_as)


@cli_job.command(name="requeue-tasks")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to requeue tasks for.")
@click.option(
    "--run-status",
    type=click.Choice(
        ["SUSPENDED", "CANCELED", "FAILED", "SUCCEEDED", "NOT_COMPATIBLE"], case_sensitive=False
    ),
    multiple=True,
    help="Requeue tasks of this status. Repeat the option to provide multiple statuses.",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@_handle_error
def job_requeue_tasks(run_status: Optional[list[str]], **args):
    """
    Requeue tasks of a [Deadline Cloud job]. By default, requeues all FAILED, CANCELED, and SUSPENDED tasks.

    Use the --run-status option to requeue tasks of different status.

    [Deadline Cloud job]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    if not run_status:
        run_status_set = {"SUSPENDED", "CANCELED", "FAILED"}
    else:
        run_status_set = {status.upper() for status in run_status}

    session = api.get_boto3_session(config=config)
    deadline_client = api.get_boto3_client("deadline", config=config)

    # Create a separate API client for the update_task calls, using the "adaptive" retry strategy.
    # This strategy is better suited to repeatedly calling the API for a longer duration, because it
    # retains backoff data across calls instead of starting fresh each time.
    requeues_client_config = get_default_client_config(
        retries=dict(mode="adaptive", total_max_attempts=5)
    )
    deadline_client_for_requeues = session.client("deadline", config=requeues_client_config)

    # Print a summary of the job's task run statuses
    job = deadline_client.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)

    click.echo(f"Job: {job['name']} ({job['jobId']})")
    # Remove the zero-count status counts for a shorter summary
    task_run_status_counts = {
        name.upper(): count for name, count in job["taskRunStatusCounts"].items() if count != 0
    }
    click.echo(_cli_object_repr({"taskRunStatusCounts": task_run_status_counts}))
    click.echo(f"Requeuing all tasks with run status among: {', '.join(sorted(run_status_set))}")

    total_task_count_to_requeue = sum(
        count for name, count in task_run_status_counts.items() if name in run_status_set
    )
    summary_task_count_by_status = ", ".join(
        f"{count} {name} tasks"
        for name, count in task_run_status_counts.items()
        if name in run_status_set and count != 0
    )
    summary_tasks_message = f"{total_task_count_to_requeue} total tasks"

    if total_task_count_to_requeue == 0:
        click.echo("No tasks to requeue.")
        return

    # Ask for confirmation about requeuing the selected tasks
    if config_file.str2bool(config_file.get_setting("settings.auto_accept", config=config)):
        click.echo(
            f"Estimated {summary_tasks_message} ({summary_task_count_by_status}) to requeue."
        )
    else:
        click.echo(
            f"This action will requeue an estimated {summary_tasks_message} ({summary_task_count_by_status})"
        )
        if not click.confirm(
            "Are you sure you want to requeue these tasks?",
            default=None,
        ):
            click.echo("No tasks were requeued.")
            sys.exit(1)
        click.echo("Requeuing tasks...")

    total_count_requeued = 0
    # Use a paginator to get all the steps
    paginator = deadline_client.get_paginator("list_steps")
    steps = []
    for page in paginator.paginate(farmId=farm_id, queueId=queue_id, jobId=job_id):
        steps.extend(page["steps"])

    # For each step, requeue all the tasks matching the run statuses
    for step in steps:
        click.echo(f"\nStep: {step['name']} ({step['stepId']})")
        step_task_count_to_requeue = sum(
            count for name, count in step["taskRunStatusCounts"].items() if name in run_status_set
        )
        summary_task_count_by_status = ", ".join(
            f"{count} {name} tasks"
            for name, count in step["taskRunStatusCounts"].items()
            if name in run_status_set and count != 0
        )
        summary_tasks_message = f"{step_task_count_to_requeue} total tasks"
        if step_task_count_to_requeue == 0:
            click.echo("  Step has no tasks to requeue.")
            continue
        else:
            click.echo(
                f"  Requeuing an estimated {summary_tasks_message} ({summary_task_count_by_status})..."
            )

        paginator = deadline_client.get_paginator("list_tasks")
        for page in paginator.paginate(
            farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step["stepId"]
        ):
            for task in page["tasks"]:
                if task["runStatus"].upper() in run_status_set:
                    parameters = task.get("parameters", {})
                    if parameters:
                        task_summary = (
                            ",".join(
                                f"{param}={list(parameters[param].values())[0]}"
                                for param in parameters
                            )
                            + f" ({task['taskId']})"
                        )
                    else:
                        task_summary = task["taskId"]
                    click.echo(f"    {task['runStatus']} {task_summary}")
                    # Requeue means to set the target run status to PENDING. If a task has no dependencies,
                    # it will switch to READY immediately.
                    deadline_client_for_requeues.update_task(
                        farmId=farm_id,
                        queueId=queue_id,
                        jobId=job_id,
                        stepId=step["stepId"],
                        taskId=task["taskId"],
                        targetRunStatus="PENDING",
                    )
                    total_count_requeued += 1

    click.echo(f"\nRequeued a total of {total_count_requeued} tasks.")


def _download_job_output(
    config: Optional[ConfigParser],
    farm_id: str,
    queue_id: str,
    job_id: str,
    step_id: Optional[str],
    task_id: Optional[str],
    is_json_format: bool = False,
):
    """
    Starts the download of job output and handles the progress reporting callback.
    """
    deadline = api.get_boto3_client("deadline", config=config)

    auto_accept = config_file.str2bool(
        config_file.get_setting("settings.auto_accept", config=config)
    )
    conflict_resolution = config_file.get_setting("settings.conflict_resolution", config=config)

    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    step = {}
    task = {}
    session_action_id = None
    if step_id:
        step = deadline.get_step(farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id)
    if task_id:
        task = deadline.get_task(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            stepId=step_id,
            taskId=task_id,
        )
        session_action_id = task.get("latestSessionActionId")

    click.echo(
        _get_start_message(job["name"], step.get("name"), task.get("parameters"), is_json_format)
    )

    queue = deadline.get_queue(farmId=farm_id, queueId=queue_id)

    queue_role_session = api.get_queue_user_boto3_session(
        deadline=deadline,
        config=config,
        farm_id=farm_id,
        queue_id=queue_id,
        queue_display_name=queue["displayName"],
    )

    # Get a dictionary mapping rootPath to rootPathFormat (OS) from job's manifests
    root_path_format_mapping: dict[str, str] = {}
    job_attachments = job.get("attachments", None)
    if job_attachments:
        job_attachments_manifests = job_attachments["manifests"]
        for manifest in job_attachments_manifests:
            root_path_format_mapping[manifest["rootPath"]] = manifest["rootPathFormat"]

    job_output_downloader = OutputDownloader(
        s3_settings=JobAttachmentS3Settings(**queue["jobAttachmentSettings"]),
        farm_id=farm_id,
        queue_id=queue_id,
        job_id=job_id,
        step_id=step_id,
        task_id=task_id,
        session_action_id=session_action_id,
        session=queue_role_session,
    )

    def _check_and_warn_long_output_paths(
        output_paths_by_root: dict[str, list[str]],
    ) -> None:
        if sys.platform == "win32" and not _is_windows_long_path_registry_enabled():
            for root, paths in output_paths_by_root.items():
                for output_path in paths:
                    if len(root + output_path) >= WINDOWS_MAX_PATH_LENGTH:
                        click.secho(
                            _get_long_path_found_message(is_json_format),
                            fg="yellow",
                        )

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()
    # If no output paths were found, log a message and exit.
    if output_paths_by_root == {}:
        click.echo(_get_no_output_message(is_json_format))
        return

    _check_and_warn_long_output_paths(output_paths_by_root)

    # Check if the asset roots came from different OS. If so, prompt users to
    # select alternative root paths to download to, (regardless of the auto-accept.)
    asset_roots = list(output_paths_by_root.keys())
    for asset_root in asset_roots:
        root_path_format = root_path_format_mapping.get(asset_root, "")
        if root_path_format == "":
            # There must be a corresponding root path format for each root path, by design.
            raise DeadlineOperationError(f"No root path format found for {asset_root}.")
        if PathFormat.get_host_path_format_string() != root_path_format:
            click.echo(_get_mismatch_os_root_warning(asset_root, root_path_format, is_json_format))

            if not is_json_format:
                new_root = click.prompt(
                    "> Please enter a new root path",
                    type=click.Path(exists=False),
                )
            else:
                json_string = click.prompt("", prompt_suffix="", type=str)
                new_root = _get_value_from_json_line(
                    json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=1
                )[0]
                _assert_valid_path(new_root)

            job_output_downloader.set_root_path(asset_root, os.path.expanduser(new_root))

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()

    _check_and_warn_long_output_paths(output_paths_by_root)

    # Prompt users to confirm local root paths where they will download outputs to,
    # and allow users to select different location to download files to if they want.
    # (If auto-accept is enabled, automatically download to the default root paths.)
    if not auto_accept:
        if not is_json_format:
            user_choice = ""
            while user_choice != ("y" or "n"):
                click.echo(
                    _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
                )
                asset_roots = list(output_paths_by_root.keys())
                click.echo(_get_roots_list_message(asset_roots, is_json_format))
                user_choice = click.prompt(
                    "> Please enter the index of root directory to edit, y to proceed without changes, or n to cancel the download",
                    type=click.Choice(
                        [
                            *[str(num) for num in list(range(0, len(asset_roots)))],
                            "y",
                            "n",
                        ]
                    ),
                    default="y",
                )
                if user_choice == "n":
                    click.echo("Output download canceled.")
                    return
                elif user_choice != "y":
                    # User selected an index to modify the root directory.
                    index_to_change = int(user_choice)
                    new_root = click.prompt(
                        "> Please enter the new root directory path, or press Enter to keep it unchanged",
                        type=click.Path(exists=False),
                        default=asset_roots[index_to_change],
                    )
                    job_output_downloader.set_root_path(
                        asset_roots[index_to_change], str(Path(new_root))
                    )
                    output_paths_by_root = job_output_downloader.get_output_paths_by_root()
                    _check_and_warn_long_output_paths(output_paths_by_root)
        else:
            click.echo(
                _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
            )
            asset_roots = list(output_paths_by_root.keys())
            click.echo(_get_roots_list_message(asset_roots, is_json_format))
            json_string = click.prompt("", prompt_suffix="", type=str)
            confirmed_asset_roots = _get_value_from_json_line(
                json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=len(asset_roots)
            )
            for index, confirmed_root in enumerate(confirmed_asset_roots):
                _assert_valid_path(confirmed_root)
                job_output_downloader.set_root_path(asset_roots[index], str(Path(confirmed_root)))
            output_paths_by_root = job_output_downloader.get_output_paths_by_root()
            _check_and_warn_long_output_paths(output_paths_by_root)

    if not is_json_format:
        # Create and print a summary of all the paths to download
        all_output_paths: set[str] = set()
        for asset_root, output_paths in output_paths_by_root.items():
            all_output_paths.update(
                os.path.normpath(os.path.join(asset_root, path)) for path in output_paths
            )
        click.echo("\nSummary of file paths to download:")
        click.echo(textwrap.indent(summarize_path_list(all_output_paths), "  "))

    # If the conflict resolution option was not specified, auto-accept is false, and
    # if there are any conflicting files in local, prompt users to select a resolution method.
    # (skip, overwrite, or make a copy.)
    if conflict_resolution != FileConflictResolution.NOT_SELECTED.name:
        file_conflict_resolution = FileConflictResolution[conflict_resolution]
    elif auto_accept:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
    else:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
        conflicting_filenames = _get_conflicting_filenames(output_paths_by_root)
        if conflicting_filenames:
            click.echo(_get_conflict_resolution_selection_message(conflicting_filenames))
            user_choice = click.prompt(
                "> Please enter your choice (1, 2, 3, or n to cancel the download)",
                type=click.Choice(["1", "2", "3", "n"]),
                default="3",
            )
            if user_choice == "n":
                click.echo("Output download canceled.")
                return
            else:
                resolution_choice_int = int(user_choice)
                file_conflict_resolution = FileConflictResolution(resolution_choice_int)

    # TODO: remove logging level setting when the max number connections for boto3 client
    # in Job Attachments library can be increased (currently using default number, 10, which
    # makes it keep logging urllib3 warning messages when downloading large files)
    with _modified_logging_level(logging.getLogger("urllib3"), logging.ERROR):

        @api.record_success_fail_telemetry_event(metric_name="download_job_output")
        def _download_job_output(
            file_conflict_resolution: Optional[
                FileConflictResolution
            ] = FileConflictResolution.CREATE_COPY,
            on_downloading_files: Optional[Callable[[ProgressReportMetadata], bool]] = None,
        ) -> DownloadSummaryStatistics:
            return job_output_downloader.download_job_output(
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=on_downloading_files,
            )

        if not is_json_format:
            # Note: click doesn't export the return type of progressbar(), so we suppress mypy warnings for
            # not annotating the type of download_progress.
            with click.progressbar(length=100, label="Downloading Outputs") as download_progress:  # type: ignore[var-annotated]

                def _update_download_progress(
                    download_metadata: ProgressReportMetadata,
                ) -> bool:
                    new_progress = int(download_metadata.progress) - download_progress.pos
                    if new_progress > 0:
                        download_progress.update(new_progress)
                    return sigint_handler.continue_operation

                download_summary: DownloadSummaryStatistics = _download_job_output(  # type: ignore
                    file_conflict_resolution=file_conflict_resolution,
                    on_downloading_files=_update_download_progress,
                )
        else:

            def _update_download_progress(
                download_metadata: ProgressReportMetadata,
            ) -> bool:
                click.echo(
                    _get_json_line(JSON_MSG_TYPE_PROGRESS, str(int(download_metadata.progress)))
                )
                # TODO: enable download cancellation for JSON format
                return True

            download_summary = _download_job_output(  # type: ignore
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=_update_download_progress,
            )

    click.echo(_get_download_summary_message(download_summary, is_json_format))
    click.echo()


def _get_start_message(
    job_name: str,
    step_name: Optional[str],
    task_parameters: Optional[dict],
    is_json_format: bool,
) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_TITLE, job_name)
    else:
        if step_name is None:
            return f"Downloading output from Job {job_name!r}"
        elif task_parameters is None:
            return f"Downloading output from Job {job_name!r} Step {step_name!r}"
        else:
            task_parameters_summary = "{}"
            if task_parameters:
                task_parameters_summary = (
                    "{"
                    + ",".join(
                        f"{key}={list(value.values())[0]}" for key, value in task_parameters.items()
                    )
                    + "}"
                )
            return f"Downloading output from Job {job_name!r} Step {step_name!r} Task {task_parameters_summary}"


def _get_no_output_message(is_json_format: bool) -> str:
    msg = (
        "There are no output files available for download at this moment. Please verify that"
        " the Job/Step/Task you are trying to download output from has completed successfully."
    )
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_SUMMARY, msg)
    else:
        return msg


def _get_mismatch_os_root_warning(root: str, root_path_format: str, is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, [root])
    else:
        path_format_capitalized_first_letter = root_path_format[0].upper() + root_path_format[1:]
        return (
            "This root path format does not match the operating system you're using. "
            "Where would you like to save the files?\n"
            f"The location was {root}, on {path_format_capitalized_first_letter}."
        )


def _get_summary_of_files_to_download_message(
    output_paths_by_root: dict[str, list[str]], is_json_format: bool
) -> str:
    # Print some information about what we will download
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PRESUMMARY, output_paths_by_root)
    else:
        paths_message_joined = "    " + "\n    ".join(
            f"{os.path.commonpath([os.path.join(directory, p) for p in output_paths])} ({len(output_paths)} file{'s' if len(output_paths) > 1 else ''})"
            for directory, output_paths in output_paths_by_root.items()
        )
        return f"\nSummary of files to download:\n{paths_message_joined}\n"


def _get_roots_list_message(asset_roots: list[str], is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, asset_roots)
    else:
        asset_roots_str = "\n".join([f"[{index}] {root}" for index, root in enumerate(asset_roots)])
        return (
            f"You are about to download files which may come from multiple root directories. Here are a list of the current root directories:\n"
            f"{asset_roots_str}"
        )


def _get_conflict_resolution_selection_message(conflicting_filenames: list[str]) -> str:
    conflicting_filenames_str = "\n        ".join(conflicting_filenames)
    return (
        f"The following files already exist in your local directory:\n        {conflicting_filenames_str}\n"
        f"You have three options to choose from:\n"
        f"[1] Skip: Do not download these files\n"
        f"[2] Overwrite: Download these files and overwrite existing files\n"
        f"[3] Create a copy: Download the file with a new name, appending '(1)' to the end"
    )


def _get_download_summary_message(
    download_summary: DownloadSummaryStatistics, is_json_format: bool
) -> str:
    if is_json_format:
        return _get_json_line(
            JSON_MSG_TYPE_SUMMARY,
            f"Downloaded {download_summary.processed_files} files",
            extra_properties={
                "fileCount": download_summary.processed_files,
                "files": download_summary.downloaded_files,
            },
        )
    else:
        paths_joined = "\n        ".join(
            f"{directory} ({count} file{'s' if count > 1 else ''})"
            for directory, count in download_summary.file_counts_by_root_directory.items()
        )
        return (
            "Download Summary:\n"
            f"    Downloaded {download_summary.processed_files} files totaling"
            f" {human_readable_file_size(download_summary.processed_bytes)}.\n"
            f"    Total download time of {round(download_summary.total_time, ndigits=5)} seconds"
            f" at {human_readable_file_size(int(download_summary.transfer_rate))}/s.\n"
            f"    Download locations (total file counts):\n        {paths_joined}"
        )


def _get_long_path_found_message(is_json_format: bool) -> str:
    message = """
WARNING: Found downloaded file paths that exceed Windows path length limit. This may cause unexpected issues.
For details and a fix using the registry, see: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
"""
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_WARNING, message)

    else:
        return message


def _get_conflicting_filenames(filenames_by_root: dict[str, list[str]]) -> list[str]:
    conflicting_filenames: list[str] = []

    for root, filenames in filenames_by_root.items():
        for filename in filenames:
            abs_path = Path(root).joinpath(filename).resolve()
            if abs_path.is_file():
                conflicting_filenames.append(str(abs_path))

    return conflicting_filenames


def _get_json_line(
    messageType: str,
    value: Union[str, list[str], dict[str, Any]],
    extra_properties: Optional[dict[str, Any]] = None,
) -> str:
    json_obj = {"messageType": messageType, "value": value}
    if extra_properties:
        json_obj.update(extra_properties)
    return json.dumps(json_obj, ensure_ascii=True)


def _get_value_from_json_line(
    json_line: str, message_type: str, expected_size: Optional[int] = None
) -> Union[str, list[str]]:
    try:
        parsed_json = json.loads(json_line)
        if parsed_json["messageType"] != message_type:
            raise ValueError(
                f"Expected message type '{message_type}' but received '{parsed_json['messageType']}'"
            )
        if expected_size and len(parsed_json["value"]) != expected_size:
            raise ValueError(
                f"Expected {expected_size} item{'' if expected_size == 1 else 's'} in value "
                f"but received {len(parsed_json['value'])}"
            )
        return parsed_json["value"]
    except Exception as e:
        raise ValueError(f"Invalid JSON line '{json_line}': {e}")


def _assert_valid_path(path: str) -> None:
    """
    Validates that the path has the format of the OS currently running.
    """
    path_obj = Path(path)
    if not path_obj.is_absolute():
        raise ValueError(f"Path {path} is not an absolute path.")


@cli_job.command(name="download-output")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to use.")
@click.option("--step-id", help="The step to use.")
@click.option("--task-id", help="The task to use.")
@click.option(
    "--conflict-resolution",
    type=click.Choice(
        [
            FileConflictResolution.SKIP.name,
            FileConflictResolution.OVERWRITE.name,
            FileConflictResolution.CREATE_COPY.name,
        ],
        case_sensitive=False,
    ),
    help="How to handle downloads if a file already exists:\n"
    "CREATE_COPY (default): Download the file with a new name, appending '(1)' to the end\n"
    "SKIP: Do not download the file\n"
    "OVERWRITE: Download and replace the existing file",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@_handle_error
def job_download_output(step_id, task_id, output, **args):
    """
    Download a the output of a [Deadline Cloud job] in the queue that was saved as [job attachments].

    [Deadline Cloud job]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    [job attachments]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/storage-job-attachments.html
    """
    if task_id and not step_id:
        raise click.UsageError("Missing option '--step-id' required with '--task-id'")
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)
    is_json_format = True if output == "json" else False

    try:
        _download_job_output(config, farm_id, queue_id, job_id, step_id, task_id, is_json_format)
    except Exception as e:
        if is_json_format:
            error_one_liner = str(e).replace("\n", ". ")
            click.echo(_get_json_line(JSON_MSG_TYPE_ERROR, error_one_liner))
            sys.exit(1)
        else:
            if logging.DEBUG >= logger.getEffectiveLevel():
                logger.exception("Exception details:")
            raise DeadlineOperationError(f"Failed to download output:\n{e}") from e


@cli_job.command(name="wait")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to wait for.")
@click.option("--max-poll-interval", default=120, help="Maximum polling interval in seconds.")
@click.option("--timeout", default=0, help="Timeout in seconds (0 for no timeout).")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format (verbose or json).",
)
@_handle_error
def job_wait_for_completion(max_poll_interval, timeout, output, **args):
    """
    Wait for a [Deadline Cloud job] to complete and then print information about failed step-task IDs.

    This command blocks until the job's taskRunStatus reaches a terminal state
    (SUCCEEDED, FAILED, CANCELED, SUSPENDED, or NOT_COMPATIBLE),
    then prints a list of any failed step-task combinations.

    The command uses exponential backoff for polling, starting at 0.5 seconds and doubling
    the interval after each check until it reaches the maximum polling interval.

    Exit codes:

    \b
        0 - Job succeeded
        1 - Timeout waiting for job completion
        2 - Job failed (any tasks failed)
        3 - Job was canceled
        4 - Job was suspended
        5 - Job is not compatible

    [Deadline Cloud job]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/deadline-cloud-jobs.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    is_json_output = output.lower() == "json"

    # Get job name for output
    deadline = api.get_boto3_client("deadline", config=config)
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job_name = job["name"]

    # Define a status callback for verbose output
    def job_callback(job, elapsed_time=0, total_timeout=0):
        if not is_json_output:
            timeout_info = ""
            if total_timeout > 0:
                remaining = max(0, total_timeout - elapsed_time)
                timeout_info = f" [{elapsed_time:.1f}s elapsed, {remaining:.1f}s remaining]"
            else:
                timeout_info = f" [{elapsed_time:.1f}s elapsed]"

            # Clear the current line and update in place
            current_worker_count = (
                job["taskRunStatusCounts"].get("RUNNING", 0)
                + job["taskRunStatusCounts"].get("ASSIGNED", 0)
                + job["taskRunStatusCounts"].get("STARTING", 0)
            )
            completed_task_count = job["taskRunStatusCounts"].get("SUCCEEDED", 0)
            total_task_count = sum(job["taskRunStatusCounts"].values())
            click.echo(
                f"\rCurrent status: {job.get('taskRunStatus', '')} ({completed_task_count}/{total_task_count} tasks succeeded, {current_worker_count} workers running).{timeout_info}",
                nl=False,
            )

    if not is_json_output:
        click.echo(f"Waiting for job {job_id} to complete...")
        click.echo(f"Job Name: {job_name}")

    try:
        result = api.wait_for_job_completion(
            farm_id=farm_id,
            queue_id=queue_id,
            job_id=job_id,
            max_poll_interval=max_poll_interval,
            timeout=timeout,
            config=config,
            job_callback=job_callback,
        )

        if is_json_output:
            # Return everything as JSON
            response = {
                "jobId": job_id,
                "jobName": job_name,
                "status": result.status,
                "elapsedTime": result.elapsed_time,
                "failedTasks": [
                    {
                        "stepId": task.step_id,
                        "taskId": task.task_id,
                        "stepName": task.step_name,
                        "sessionId": task.session_id,
                    }
                    for task in result.failed_tasks
                ],
            }
            click.echo(json.dumps(response, indent=2))
        else:
            # Use verbose output with YAML formatting
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job completed with status: {result.status}")
            click.echo(f"Elapsed time: {result.elapsed_time:.1f} seconds")

            if result.failed_tasks:
                click.echo(f"Found {len(result.failed_tasks)} failed tasks:")
                failed_tasks_dict = [
                    {
                        "stepId": task.step_id,
                        "taskId": task.task_id,
                        "stepName": task.step_name,
                        "sessionId": task.session_id,
                    }
                    for task in result.failed_tasks
                ]
                click.echo(_cli_object_repr(failed_tasks_dict))
            else:
                click.echo("No failed tasks found.")

        # Determine exit code based on job status
        exit_code = 0
        if result.status == "SUCCEEDED" and not result.failed_tasks:
            exit_code = 0
        elif result.status == "FAILED" or result.failed_tasks:
            exit_code = 2
        elif result.status == "CANCELED":
            exit_code = 3
        elif result.status == "SUSPENDED":
            exit_code = 4
        elif result.status == "ARCHIVED":
            exit_code = 4
        elif result.status == "NOT_COMPATIBLE":
            exit_code = 5
        else:
            # Unknown status, treat as failure
            exit_code = 2

        sys.exit(exit_code)

    except DeadlineOperationTimedOut as e:
        if is_json_output:
            error_response = {
                "error": str(e),
                "timeout": True,
                "jobId": job_id,
                "jobName": job_name,
            }
            click.echo(json.dumps(error_response, indent=2))
        else:
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")
            click.echo(f"Error waiting for job completion: {e}")
        sys.exit(1)
    except DeadlineOperationError as e:
        if is_json_output:
            error_response = {
                "error": str(e),
                "jobId": job_id,
                "jobName": job_name,
            }
            click.echo(json.dumps(error_response, indent=2))
        else:
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")
            click.echo(f"Error waiting for job completion: {e}")
        sys.exit(2)


@cli_job.command(name="logs")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to get logs for.")
@click.option(
    "--session-id",
    help="The session ID to get logs for. If not provided and job-id is specified, will use the latest session based on endedAt time.",
)
@click.option(
    "--session-action-id",
    help="The session action ID to get logs for. The session ID will be derived from this.",
)
@click.option("--limit", default=100, help="Maximum number of log lines to return.")
@click.option(
    "--start-time",
    help="Start time for logs in ISO format (e.g., 2023-01-01T12:00:00Z).",
)
@click.option("--end-time", help="End time for logs in ISO format (e.g., 2023-01-01T13:00:00Z).")
@click.option("--next-token", help="Token for pagination of results.")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format (verbose or json).",
)
@click.option(
    "--timestamp-format",
    type=click.Choice(["utc", "local", "relative"], case_sensitive=False),
    default="utc",
    help="Timestamp format for log entries (utc, local, or relative). Default is utc.",
)
@click.option(
    "--timezone",
    type=click.Choice(["utc", "local"], case_sensitive=False),
    help="[DEPRECATED] Use --timestamp-format instead. Timezone for timestamps (utc or local).",
)
@click.pass_context
@_handle_error
def job_logs(
    ctx,
    session_id,
    session_action_id,
    limit,
    start_time,
    end_time,
    next_token,
    output,
    timestamp_format,
    timezone,
    **args,
):
    """
    Prints a [Deadline Cloud session log] stored in [CloudWatch logs] for the job.
    Defaults to a recent/ongoing session if a session id is not provided.

    By default, it returns the most recent 100 log lines, but this can be
    adjusted using the --limit parameter.

    If session-id is not provided but job-id is, the command will automatically
    select a session using the following priority:
    1. If there are ongoing sessions (no endedAt time), always prefer them
    2. Among ongoing sessions, select the one that started most recently
    3. If no ongoing sessions exist, select the completed session that ended most recently

    Use --next-token with the value from a previous response to get the next page of results
    of log output prior to the last page.

    [Deadline Cloud session log]: https://docs.aws.amazon.com/deadline-cloud/latest/userguide/view-logs.html
    [CloudWatch logs]: https://docs.aws.amazon.com/deadline-cloud/latest/developerguide/monitoring-cloudwatch.html
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    is_json_output = output.lower() == "json"

    # Check if --timestamp-format was explicitly provided by the user
    timestamp_format_provided = (
        ctx.get_parameter_source("timestamp_format") != click.core.ParameterSource.DEFAULT
    )

    # Handle timestamp format options and deprecation
    if timezone and timestamp_format_provided:
        raise DeadlineOperationError(
            "Cannot use both --timezone and --timestamp-format options. Use --timestamp-format instead."
        )

    # Handle deprecation of --timezone option
    if timezone:
        # Only show deprecation warning for non-JSON output to avoid interfering with JSON parsing
        if not is_json_output:
            warning_message = (
                f"Warning: --timezone is deprecated and will be removed in a future version. "
                f"Use --timestamp-format {timezone} instead."
            )
            click.echo(warning_message, err=True)
        # Map deprecated timezone values to new format
        timestamp_format = timezone

    # Convert timestamp_format to enum
    timestamp_format_enum = TimestampFormat(timestamp_format.lower())

    # Handle session action ID if provided
    if session_action_id:
        # Parse session action ID to derive session ID
        derived_session_id = _parse_session_action_id(session_action_id)

        # If both session_id and session_action_id are provided, validate consistency
        if session_id:
            if session_id != derived_session_id:
                raise DeadlineOperationError(
                    f"Session ID mismatch: --session-id '{session_id}' does not match "
                    f"session ID '{derived_session_id}' derived from --session-action-id '{session_action_id}'"
                )
        else:
            # If only session_action_id is provided, set session_id from parsed value
            session_id = derived_session_id

    # Get job name if job_id is available
    deadline = api.get_boto3_client("deadline", config=config)
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job_name = job["name"]

    # If session_id is not provided but job_id is, try to find the session
    if not session_id and job_id:
        try:
            # Use paginator to get all sessions
            paginator = deadline.get_paginator("list_sessions")
            sessions = []

            for page in paginator.paginate(farmId=farm_id, queueId=queue_id, jobId=job_id):
                sessions.extend(page.get("sessions", []))

            if not sessions:
                raise DeadlineOperationError(f"No sessions found for job {job_id}")
            elif len(sessions) == 1:
                session_id = sessions[0]["sessionId"]
                if not is_json_output:
                    click.echo(f"Using the only available session: {session_id}")
            else:
                # Multiple sessions found, select the latest one
                # Prioritize ongoing sessions (no endedAt) over completed ones
                # Among ongoing sessions, select the one that started most recently
                # Among completed sessions, select the one that ended most recently
                ongoing_sessions = [s for s in sessions if "endedAt" not in s]
                completed_sessions = [s for s in sessions if "endedAt" in s]

                if ongoing_sessions:
                    # Always prefer ongoing sessions, select the most recently started one
                    latest_session = max(
                        ongoing_sessions,
                        key=lambda s: s.get(
                            "startedAt", datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
                        ),
                    )
                else:
                    # No ongoing sessions, select the most recently completed one
                    latest_session = max(
                        completed_sessions,
                        key=lambda s: s.get(
                            "endedAt", datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
                        ),
                    )

                session_id = latest_session["sessionId"]
                if not is_json_output:
                    click.echo(f"Using the latest session: {session_id}")
        except ClientError as exc:
            raise DeadlineOperationError(
                f"Failed to list sessions for job {job_id}:\n{exc}"
            ) from exc

    # Ensure we have a session ID at this point
    if not session_id:
        raise DeadlineOperationError(
            "Session ID is required. Provide it with --session-id or specify a --job-id with at least one session."
        )

    if session_action_id:
        log_entity = f"session action {session_action_id}"
    else:
        log_entity = f"session {session_id}"

    if not is_json_output:
        click.echo(
            f"Retrieving logs for {log_entity} from log group /aws/deadline/{farm_id}/{queue_id}..."
        )
        # Display job information if available
        click.echo(f"Job ID: {job_id}")
        click.echo(f"Job Name: {job_name}")

    # If session_action_id is provided, retrieve session action details to get timestamp range
    if session_action_id:
        try:
            session_action = deadline.get_session_action(
                farmId=farm_id,
                queueId=queue_id,
                jobId=job_id,
                sessionActionId=session_action_id,
            )
        except ClientError as exc:
            if exc.response.get("Error", {}).get("Code") == "ResourceNotFoundException":
                raise DeadlineOperationError(
                    f"Session action '{session_action_id}' not found in job '{job_id}'"
                ) from exc
            else:
                raise DeadlineOperationError(
                    f"Failed to retrieve session action '{session_action_id}':\n{exc}"
                ) from exc

        # Extract timestamps from session action
        action_start = session_action.get("startedAt")
        action_end = session_action.get("endedAt")

        # If session action hasn't started yet, raise an error
        if not action_start:
            raise DeadlineOperationError(
                f"Session action '{session_action_id}' has not started yet. No logs are available."
            )

        # Set reference time for relative format (session action start time)
        reference_start_time = action_start

        if not is_json_output:
            click.echo(f"Session action start: {action_start}")
        if action_end:
            if not is_json_output:
                click.echo(f"Session action end: {action_end}")
        else:
            if not is_json_output:
                click.echo("Session action is still running")
            # If session action is still in progress, use current time as end time
            action_end = datetime.datetime.now(datetime.timezone.utc)
        if not is_json_output:
            click.echo(f"Session action duration: {action_end - action_start}")

        # Get the effective time range to process, which is the intersection of
        # the session action time range and the user-provided time range
        if start_time:
            # Parse user's start_time
            user_start = datetime.datetime.fromisoformat(start_time.replace("Z", "+00:00"))
            # Use the later of the two start times
            effective_start = max(user_start, action_start)
        else:
            effective_start = action_start

        if end_time:
            # Parse user's end_time
            user_end = datetime.datetime.fromisoformat(end_time.replace("Z", "+00:00"))
            # Use the earlier of the two end times
            effective_end = min(user_end, action_end)
        else:
            effective_end = action_end

        # Validate that effective_start <= effective_end
        if effective_start > effective_end:
            error_msg = (
                f"The specified time range does not overlap with the session action's time range.\n"
                f"Session action time range: {action_start.isoformat()} to {action_end.isoformat()}"
            )
            error_msg += f"\nSpecified time range: {start_time or 'N/A'} to {end_time or 'N/A'}"
            raise DeadlineOperationError(error_msg)

        # Use the combined timestamps for log retrieval
        start_time = effective_start
        end_time = effective_end
    else:
        # Get session start time for the timestamp formatter
        try:
            session_details = deadline.get_session(
                farmId=farm_id, queueId=queue_id, jobId=job_id, sessionId=session_id
            )
            try:
                reference_start_time = session_details["startedAt"]
            except KeyError:
                raise DeadlineOperationError(
                    f"Session '{session_id}' has not started yet."
                ) from None
        except ClientError as exc:
            raise DeadlineOperationError(
                f"Failed to retrieve session details for session ID '{session_id}':\n{exc}"
            ) from exc

    # Create timestamp formatter now that we have all necessary information
    timestamp_formatter = TimestampFormatter(timestamp_format_enum, reference_start_time)

    try:
        result = api.get_session_logs(
            farm_id=farm_id,
            queue_id=queue_id,
            session_id=session_id,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            next_token=next_token,
            config=config,
        )

        if is_json_output:
            # Return everything as JSON
            response = {
                "jobId": job_id,
                "jobName": job_name,
                "events": [
                    {
                        "timestamp": timestamp_formatter.format_timestamp(event.timestamp),
                        "message": event.message,
                        "ingestionTime": (
                            timestamp_formatter.format_timestamp(event.ingestion_time)
                            if event.ingestion_time
                            else None
                        ),
                        "eventId": event.event_id,
                    }
                    for event in result.events
                ],
                "count": result.count,
                "nextToken": result.next_token,
                "logGroup": result.log_group,
                "logStream": result.log_stream,
            }
            click.echo(json.dumps(response, indent=2))
        else:
            # Display reference time for relative format
            if (
                timestamp_format_enum == TimestampFormat.RELATIVE
                and timestamp_formatter.reference_start_time
            ):
                reference_display = timestamp_formatter.reference_start_time.isoformat()
                click.echo(f"Logs relative to start time: {reference_display}")

            # Separate the logs by a blank line to make the start of the log easy to find.
            click.echo()

            # Display the logs in verbose format
            if not result.events:
                click.echo("No logs found for the specified session.")
                return

            for event in result.events:
                timestamp = timestamp_formatter.format_timestamp(event.timestamp)
                click.echo(f"[{timestamp}] {event.message}")

            click.echo(f"\nRetrieved {result.count} log events.")

            # If there are more logs available, inform the user
            if result.next_token:
                click.echo(
                    f'More logs are available. Use --next-token "{result.next_token}" to retrieve the next page.'
                )

    except Exception as e:
        if is_json_output:
            error_response = {"error": str(e)}
            click.echo(json.dumps(error_response, indent=2))
            sys.exit(1)
        else:
            raise DeadlineOperationError(f"Error retrieving logs: {e}")


@cli_job.command(name="trace-schedule")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to trace.")
@click.option("-v", "--verbose", is_flag=True, help="Output verbose trace details.")
@click.option(
    "--trace-format",
    type=click.Choice(
        ["chrome"],
        case_sensitive=False,
    ),
    help="The tracing format to write.",
)
@click.option("--trace-file", help="The tracing file to write.")
@_handle_error
def job_trace_schedule(verbose, trace_format, trace_file, **args):
    """
    EXPERIMENTAL - Generate statistics from a job with a trace that you can view and explore interactively.

    To visualize the trace output file when providing the options
    "--trace-format chrome --trace-file output.json", open
    the [Perfetto Tracing UI] in a browser and choose "Open trace file".

    [Perfetto Tracing UI]: https://ui.perfetto.dev
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    if trace_file and not trace_format:
        raise DeadlineOperationError("Error: Must provide --trace-format with --trace-file.")

    deadline = api.get_boto3_client("deadline", config=config)
    trace_end_utc = datetime.datetime.now(datetime.timezone.utc)

    click.echo("Getting the job...")
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job.pop("ResponseMetadata", None)

    if "startedAt" not in job:
        raise DeadlineOperationError("No trace available - Job hasn't started yet, exiting")
    started_at = job["startedAt"]

    click.echo("Getting all the sessions for the job...")
    response = deadline.list_sessions(farmId=farm_id, queueId=queue_id, jobId=job_id)
    while "nextToken" in response:
        old_list = response["sessions"]
        response = deadline.list_sessions(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            nextToken=response["nextToken"],
        )
        response["sessions"] = old_list + response["sessions"]
    response.pop("ResponseMetadata", None)

    sessions = sorted(response["sessions"], key=lambda session: session["startedAt"])

    click.echo("Getting all the session actions for the job...")
    for session in sessions:
        response = deadline.list_session_actions(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            sessionId=session["sessionId"],
        )
        while "nextToken" in response:
            old_list = response["sessionActions"]
            response = deadline.list_session_actions(
                farmId=farm_id,
                queueId=queue_id,
                jobId=job_id,
                sessionId=session["sessionId"],
                nextToken=response["nextToken"],
            )
            response["sessionActions"] = old_list + response["sessionActions"]
        response.pop("ResponseMetadata", None)

        session["actions"] = response["sessionActions"]

    # Cache steps and tasks by their id, to only get each once
    steps: dict[str, Any] = {}
    tasks: dict[str, Any] = {}

    with click.progressbar(  # type: ignore[var-annotated]
        length=len(sessions), label="Getting all the steps and tasks for the job..."
    ) as progressbar:
        for index, session in enumerate(sessions):
            session["index"] = index
            for action in session["actions"]:
                step_id = action["definition"].get("taskRun", {}).get("stepId")
                task_id = action["definition"].get("taskRun", {}).get("taskId")
                if step_id and task_id:
                    if "step" not in session:
                        if step_id in steps:
                            step = steps[step_id]
                        else:
                            step = deadline.get_step(
                                farmId=farm_id,
                                queueId=queue_id,
                                jobId=job_id,
                                stepId=step_id,
                            )
                            step.pop("ResponseMetadata", None)
                            steps[step_id] = step
                        session["step"] = step
                    elif session["step"]["stepId"] != step_id:
                        # The session itself doesn't have a step id, but for now the scheduler always creates new
                        # sessions for new steps.
                        raise DeadlineOperationError(
                            f"Session {session['sessionId']} ran more than one step! When this code was"
                            " written that wasn't possible."
                        )

                    if task_id in tasks:
                        task = tasks[task_id]
                    else:
                        task = deadline.get_task(
                            farmId=farm_id,
                            queueId=queue_id,
                            jobId=job_id,
                            stepId=step_id,
                            taskId=task_id,
                        )
                        task.pop("ResponseMetadata", None)
                        tasks[task_id] = task
                    action["task"] = task
            progressbar.update(1)

    # Collect the worker IDs that ran the sessions, and give them indexes to act as PIDs in the tracing file
    worker_ids = {session["workerId"] for session in sessions}
    workers = {worker_id: index for index, worker_id in enumerate(worker_ids)}

    click.echo("Processing the trace data...")
    trace_events = []

    def time_int(timestamp: datetime.datetime):
        return int((timestamp - started_at) / datetime.timedelta(microseconds=1))

    def duration_of(resource):
        try:
            return time_int(resource.get("endedAt", trace_end_utc)) - time_int(
                resource["startedAt"]
            )
        except KeyError:
            return 0

    accumulators = {
        "sessionCount": 0,
        "sessionActionCount": 0,
        "taskRunCount": 0,
        "envActionCount": 0,
        "syncJobAttachmentsCount": 0,
        "sessionDuration": 0,
        "sessionActionDuration": 0,
        "taskRunDuration": 0,
        "envActionDuration": 0,
        "syncJobAttachmentsDuration": 0,
    }

    for session in sessions:
        accumulators["sessionCount"] += 1
        accumulators["sessionDuration"] += duration_of(session)

        pid = workers[session["workerId"]]
        session_event_name = f"{session['step']['name']} - {session['index']}"
        if "endedAt" not in session:
            session_event_name = f"{session_event_name} - In Progress"
        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "B",  # Begin Event
                "ts": time_int(session["startedAt"]),
                "pid": pid,
                "tid": 0,
                "args": {
                    "sessionId": session["sessionId"],
                    "workerId": session["workerId"],
                    "fleetId": session["fleetId"],
                    "lifecycleStatus": session["lifecycleStatus"],
                },
            }
        )

        for action in session["actions"]:
            accumulators["sessionActionCount"] += 1
            accumulators["sessionActionDuration"] += duration_of(action)

            name = action["sessionActionId"]
            action_type = list(action["definition"].keys())[0]
            if action_type == "taskRun":
                accumulators["taskRunCount"] += 1
                accumulators["taskRunDuration"] += duration_of(action)

                task = action["task"]
                parameters = task.get("parameters", {})
                name = ",".join(
                    f"{param}={list(parameters[param].values())[0]}" for param in parameters
                )
                if not name:
                    name = "<No Task Params>"
            elif action_type in ("envEnter", "envExit"):
                accumulators["envActionCount"] += 1
                accumulators["envActionDuration"] += duration_of(action)

                name = action["definition"][action_type]["environmentId"].split(":")[-1]
            elif action_type == "syncInputJobAttachments":
                accumulators["syncJobAttachmentsCount"] += 1
                accumulators["syncJobAttachmentsDuration"] += duration_of(action)

                if "stepId" in action["definition"][action_type]:
                    name = "Sync Job Attchmnt (Dependencies)"
                else:
                    name = "Sync Job Attchmnt (Submitted)"
            if "endedAt" not in action:
                name = f"{name} - In Progress"
            if "startedAt" in action:
                trace_events.append(
                    {
                        "name": name,
                        "cat": action_type,
                        "ph": "X",  # Complete Event
                        "ts": time_int(action["startedAt"]),
                        "dur": duration_of(action),
                        "pid": pid,
                        "tid": 0,
                        "args": {
                            "sessionActionId": action["sessionActionId"],
                            "status": action["status"],
                            "stepName": session["step"]["name"],
                        },
                    }
                )

        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "E",  # End Event
                "ts": time_int(session.get("endedAt", trace_end_utc)),
                "pid": pid,
                "tid": 0,
            }
        )

    if verbose:
        click.echo(" ==== TRACE DATA ====")
        click.echo(_cli_object_repr(job))
        click.echo("")
        click.echo(_cli_object_repr(sessions))

    click.echo("")
    click.echo(" ==== SUMMARY ====")
    click.echo("")
    click.echo(f"Session Count: {accumulators['sessionCount']}")
    session_total_duration = accumulators["sessionDuration"]
    click.echo(f"Session Total Duration: {datetime.timedelta(microseconds=session_total_duration)}")
    click.echo(f"Session Action Count: {accumulators['sessionActionCount']}")
    click.echo(
        f"Session Action Total Duration: {datetime.timedelta(microseconds=accumulators['sessionActionDuration'])}"
    )
    click.echo(f"Task Run Count: {accumulators['taskRunCount']}")
    task_run_total_duration = accumulators["taskRunDuration"]
    click.echo(
        f"Task Run Total Duration: {datetime.timedelta(microseconds=task_run_total_duration)} ({100 * task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Non-Task Run Count: {accumulators['sessionActionCount'] - accumulators['taskRunCount']}"
    )
    non_task_run_total_duration = (
        accumulators["sessionActionDuration"] - accumulators["taskRunDuration"]
    )
    click.echo(
        f"Non-Task Run Total Duration: {datetime.timedelta(microseconds=non_task_run_total_duration)} ({100 * non_task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Sync Job Attachments Count: {accumulators['syncJobAttachmentsCount']}")
    sync_job_attachments_total_duration = accumulators["syncJobAttachmentsDuration"]
    click.echo(
        f"Sync Job Attachments Total Duration: {datetime.timedelta(microseconds=sync_job_attachments_total_duration)} ({100 * sync_job_attachments_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Env Action Count: {accumulators['envActionCount']}")
    env_action_total_duration = accumulators["envActionDuration"]
    click.echo(
        f"Env Action Total Duration: {datetime.timedelta(microseconds=env_action_total_duration)} ({100 * env_action_total_duration / session_total_duration:.1f}%)"
    )
    click.echo("")
    within_session_overhead_duration = (
        accumulators["sessionDuration"] - accumulators["sessionActionDuration"]
    )
    click.echo(
        f"Within-session Overhead Duration: {datetime.timedelta(microseconds=within_session_overhead_duration)} ({100 * within_session_overhead_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Within-session Overhead Duration Per Action: {datetime.timedelta(microseconds=(accumulators['sessionDuration'] - accumulators['sessionActionDuration']) / accumulators['sessionActionCount'])}"
    )

    tracing_data: dict[str, Any] = {
        "traceEvents": trace_events,
        # "displayTimeUnits": "s",
        "otherData": {
            "farmId": farm_id,
            "queueId": queue_id,
            "jobId": job_id,
            "jobName": job["name"],
            "startedAt": job["startedAt"].isoformat(sep="T"),
        },
    }
    if "endedAt" in job:
        tracing_data["otherData"]["endedAt"] = job["endedAt"].isoformat(sep="T")

    tracing_data["otherData"].update(accumulators)

    if trace_file:
        with open(trace_file, "w", encoding="utf8") as f:
            json.dump(tracing_data, f, indent=1)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

import os
import shutil

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from typing import Any
import json


class HatchCustomBuildHook(BuildHookInterface):
    """
    This class implements Hatch's [custom build hook] (https://hatch.pypa.io/1.6/plugins/build-hook/custom/)
    for a copy_version_py operation that copies the _version.py file generated by the hatch-vcs build hook into
    specified destination directories. See the `[[tool.hatch.build.hooks.custom]]` section in `pyproject.toml`.
    """

    def _validate_config(self):
        if sorted(self.config) != ["copy_version_py", "path"] or list(
            self.config["copy_version_py"]
        ) != ["destinations"]:
            raise RuntimeError(
                "Configuration of the custom build hook must be like { 'copy_version_py': {'destinations': ['path1', ...]}}."
                + f" Received:\n{self.config}"
            )

    def _compile_translations(self):
        """
        Generates types for translation strings based on the English translation JSON file. The types
        give feedback during development and validation during linting that all translated strings are
        part of the translation files.
        """
        translations_dir = os.path.join(
            self.root, "src", "deadline", "client", "ui", "translations", "locales"
        )

        # Load en_US as the source for type generation
        en_file = os.path.join(translations_dir, "en_US.json")
        with open(en_file) as f:
            translations = json.load(f)

        # Generate type hints file
        keys = list(translations.keys())
        # Escape for Python string literals: backslashes, quotes, newlines
        escaped_keys = [
            k.replace("\\", "\\\\").replace('"', '\\"').replace("\n", "\\n") for k in keys
        ]
        type_file = os.path.join(
            self.root, "src", "deadline", "client", "ui", "_translation_keys.py"
        )

        with open(type_file, "w", encoding="utf-8") as f:
            f.write("# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.\n")
            f.write("# Auto-generated from locales/en_US.json - DO NOT EDIT\n\n")
            f.write("from typing import Literal\n\n")
            f.write("TranslationKey = Literal[\n")
            f.write(",\n".join(f'    "{k}"' for k in escaped_keys))
            f.write("\n]\n")

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        self._validate_config()
        self._compile_translations()

        for destination in self.config["copy_version_py"]["destinations"]:
            print(f"Copying _version.py to {destination}")
            shutil.copy(
                os.path.join(self.root, "_version.py"),
                os.path.join(self.root, destination),
            )

    def clean(self, versions: list[str]) -> None:
        self._validate_config()

        cleaned_count = 0
        for destination in self.config["copy_version_py"]["destinations"]:
            print(f"Cleaning _version.py from {destination}")
            clean_path = os.path.join(self.root, destination, "_version.py")
            try:
                os.remove(clean_path)
                cleaned_count += 1
            except FileNotFoundError:
                pass
        print(f"Cleaned {cleaned_count} items")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
A UI Widget containing the timeout settings widget.
"""

from __future__ import annotations

from typing import Dict, Optional
from datetime import timedelta

from qtpy.QtWidgets import (  # type: ignore
    QFormLayout,
    QGroupBox,
    QLabel,
    QSpinBox,
    QCheckBox,
    QGridLayout,
    QWidget,
)
from qtpy.QtCore import Signal
from .._utils import tr
from ..dataclasses.timeouts import TimeoutTableEntries

# UI Constants
WARNING_ICON = "⚠️"
ERROR_ICON = "❌"
ERROR_BG_COLOR = "#FFE4E1"  # Light red
WARNING_BG_COLOR = "#FFF3CD"  # Light yellow


class TimeoutEntryWidget(QWidget):
    """
    A widget representing a single timeout row with checkbox, status icon, and time input fields.

    Contains all the UI elements and logic for managing a single timeout entry including
    checkbox state, validation indicators, and time value calculations.
    """

    # Signal to indicate any change in the widget
    changed = Signal()

    def __init__(self, label: str, tooltip: str):
        super().__init__()

        self.checkbox = QCheckBox(label)
        self.checkbox.setChecked(True)
        self.checkbox.setToolTip(tooltip)

        self.status_icon = QLabel()
        self.status_icon.setFixedSize(16, 16)

        self.days_box = QSpinBox(self, minimum=0, maximum=365)
        self.days_box.setSuffix(" days")
        self.days_box.setFixedWidth(90)

        self.hours_box = QSpinBox(self, minimum=0, maximum=23)
        self.hours_box.setSuffix(" hours")
        self.hours_box.setFixedWidth(90)

        self.minutes_box = QSpinBox(self, minimum=0, maximum=59)
        self.minutes_box.setSuffix(" minutes")
        self.minutes_box.setFixedWidth(90)

        self._build_ui()
        self._connect_signals()

    def _build_ui(self):
        """
        Builds the internal layout of the timeout row.
        """
        layout = QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(self.checkbox, 0, 0)
        layout.addWidget(self.status_icon, 0, 1)
        layout.addWidget(self.days_box, 0, 2)
        layout.addWidget(self.hours_box, 0, 3)
        layout.addWidget(self.minutes_box, 0, 4)

    def _connect_signals(self):
        """
        Connects all internal widgets to emit the changed signal.
        """
        self.checkbox.clicked.connect(self._on_change)
        self.days_box.valueChanged.connect(self._on_change)
        self.hours_box.valueChanged.connect(self._on_change)
        self.minutes_box.valueChanged.connect(self._on_change)

    def _on_change(self):
        """
        Handler for any changes in the widget's state.
        Updates the widget state and emits the changed signal.
        """
        self.update_state()
        self.changed.emit()

    def set_enabled(self, enabled: bool):
        """
        Enables or deactivates all time input fields in this row.

        Args:
            enabled: If True, enables the input fields; if False, deactivates them
        """
        self.days_box.setEnabled(enabled)
        self.hours_box.setEnabled(enabled)
        self.minutes_box.setEnabled(enabled)

    def set_timeout(self, total_seconds: int) -> None:
        """
        Sets the timeout value by converting seconds into days, hours, and minutes.

        Args:
            total_seconds: Total number of seconds to be distributed across time fields
        """
        td = timedelta(seconds=total_seconds)

        # Extract days from timedelta (1 day = 86400 seconds)
        days = td.days

        # Extract hours from remaining seconds (td.seconds is always < 86400)
        hours = td.seconds // 3600  # 3600 = 60 minutes * 60 seconds

        # Extract minutes from remaining seconds after hours are removed
        minutes = (td.seconds // 60) % 60

        self.days_box.setValue(days)
        self.hours_box.setValue(hours)
        self.minutes_box.setValue(minutes)

    def get_timeout_seconds(self) -> int:
        """
        Calculates the total timeout in seconds from the current days, hours, and minutes values.

        Returns:
            int: Total number of seconds represented by the current time values
        """
        return int(
            timedelta(
                days=self.days_box.value(),
                hours=self.hours_box.value(),
                minutes=self.minutes_box.value(),
            ).total_seconds()
        )

    def update_suffix(self):
        """
        Updates the suffix of time input fields to be singular or plural based on their values.
        For example: "1 day" vs "2 days"
        """
        for box in [self.days_box, self.hours_box, self.minutes_box]:
            suffix = box.suffix().rstrip("s")
            box.setSuffix(suffix if box.value() == 1 else suffix + "s")

    def set_status_icon(self, status: str):
        """
        Sets the status icon (warning or error) next to the checkbox.

        Args:
            status: The icon to display (WARNING_ICON, ERROR_ICON, or empty string)
        """
        self.status_icon.setText(status)

    def set_error_style(self, is_error: bool):
        """
        Sets the error styling for the time input fields.

        Args:
            is_error: If True, applies error styling; if False, removes it
        """
        style = f"QSpinBox {{ background-color: {ERROR_BG_COLOR}; }}" if is_error else ""
        for box in [self.days_box, self.hours_box, self.minutes_box]:
            box.setStyleSheet(style)

    def update_state(self):
        """
        Updates the complete state of the row including enabled state, validation indicators,
        and suffixes based on current values and checkbox state.
        """
        is_checked = self.checkbox.isChecked()
        if not is_checked:
            self.set_status_icon(WARNING_ICON)
            self.set_error_style(False)
        elif self.get_timeout_seconds() == 0:
            self.set_status_icon(ERROR_ICON)
            self.set_error_style(True)
        else:
            self.set_status_icon("")
            self.set_error_style(False)
        self.set_enabled(is_checked)
        self.update_suffix()


class TimeoutTableWidget(QGroupBox):
    """
    A widget for managing multiple timeout settings in a tabular format.

    This widget provides a user interface for configuring multiple timeout entries,
    each consisting of a checkbox for activation/deactivation and time input fields
    for days, hours, and minutes. It includes validation, warning indicators, and
    error messages for invalid configurations.

    Features:
    - Individual timeout rows with checkbox activation
    - Time input fields for days, hours, and minutes
    - Visual indicators for warnings and errors
    - Real-time validation and feedback
    - Automatic suffix updates (singular/plural)
    - Comprehensive error messaging

    Args:
        timeouts (TimeoutTableEntries): Configuration object containing timeout entries
        parent (QWidget, optional): Parent widget. Defaults to None.
    """

    def __init__(self, *, timeouts: TimeoutTableEntries, parent: Optional[QWidget] = None):
        super().__init__(tr("Timeouts"), parent=parent)
        self.timeout_rows: Dict[str, TimeoutEntryWidget] = {}
        self._build_ui(timeouts)
        self.refresh_ui(timeouts)

    def _build_ui(self, timeouts: TimeoutTableEntries):
        """
        Constructs the complete UI layout with all timeout rows and message labels.

        Args:
            entries: List of TimeoutEntry objects defining the timeout settings
        """
        self.layout = QFormLayout(self)
        self.layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)

        self.timeouts_box = QWidget()
        timeouts_layout = QGridLayout(self.timeouts_box)

        for index, (label, entry) in enumerate(timeouts.entries.items()):
            timeout_row = TimeoutEntryWidget(label, entry.tooltip)
            timeouts_layout.addWidget(timeout_row, index, 0, 1, 4)
            self.timeout_rows[label] = timeout_row
            timeout_row.changed.connect(self._update_ui_state)

        self.error_label = self._create_message_label(ERROR_BG_COLOR)
        self.warning_label = self._create_message_label(WARNING_BG_COLOR)

        timeouts_layout.addWidget(self.error_label, len(timeouts.entries), 0, 1, 4)
        timeouts_layout.addWidget(self.warning_label, len(timeouts.entries) + 1, 0, 1, 4)

        self.layout.addRow(self.timeouts_box)

    def _create_message_label(self, bg_color: str) -> QLabel:
        """
        Creates a styled label for displaying error or warning messages.

        Args:
            bg_color: Background color for the message label

        Returns:
            QLabel: A configured label with appropriate styling
        """
        label = QLabel()
        label.setStyleSheet(f"""
            QLabel {{
                background-color: {bg_color};
                color: black;
                padding: 5px;
                border-radius: 3px;
            }}
        """)
        label.setWordWrap(True)
        return label

    def _update_ui_state(self):
        """
        Updates the UI state for error and warning messages.
        """
        self._update_error_states()
        self._update_warning_states()

    def _update_error_states(self):
        """
        Updates the error message label based on validation of all timeout rows.
        """
        any_zero = any(
            row.get_timeout_seconds() == 0 and row.checkbox.isChecked()
            for row in self.timeout_rows.values()
        )
        self.error_label.setText(tr("Error: Timeout cannot be set to zero.") if any_zero else "")
        self.error_label.setVisible(any_zero)

    def _update_warning_states(self):
        """
        Updates the warning message label based on the state of all timeout rows.
        """
        any_deactivated = any(not row.checkbox.isChecked() for row in self.timeout_rows.values())
        warning_text = (
            "Warning: Without a specified timeout, tasks may run indefinitely if issues occur."
            if any_deactivated
            else ""
        )
        self.warning_label.setText(warning_text)
        self.warning_label.setVisible(any_deactivated)

    def refresh_ui(self, timeouts: TimeoutTableEntries):
        """
        Refreshes all UI elements to reflect the current timeout settings.

        Args:
            timeouts: List of TimeoutEntry objects containing the current settings
        """
        for label, entry in timeouts.entries.items():
            if label in self.timeout_rows:
                row = self.timeout_rows[label]
                row.checkbox.setChecked(entry.is_activated)
                row.set_timeout(entry.seconds)
        self._update_ui_state()

    def update_settings(self, timeouts: TimeoutTableEntries):
        """
        Updates the timeouts with the current values from the UI.
        """
        for label, row in self.timeout_rows.items():
            if label in timeouts.entries:
                entry = timeouts.entries[label]
                entry.is_activated = row.checkbox.isChecked()
                entry.seconds = row.get_timeout_seconds()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import boto3
import json

from typing import Any, Optional, List, Dict, Callable
from pathlib import Path

from deadline.job_attachments.api._utils import _read_manifests
from deadline.job_attachments.asset_manifests.base_manifest import BaseAssetManifest
from deadline.job_attachments.download import download_files_from_manifests
from deadline.job_attachments.models import (
    FileConflictResolution,
    JobAttachmentS3Settings,
    UploadManifestInfo,
    PathMappingRule,
)
from deadline.job_attachments.progress_tracker import DownloadSummaryStatistics
from deadline.job_attachments.upload import S3AssetUploader

from deadline.client.config import config_file
from deadline.client.exceptions import NonValidInputError


def _attachment_download(
    manifests: List[str],
    s3_root_uri: str,
    boto3_session: boto3.Session,
    path_mapping_rules: Optional[str] = None,
    print_function_callback: Callable[[Any], None] = lambda msg: None,
    conflict_resolution: FileConflictResolution = FileConflictResolution.CREATE_COPY,
) -> DownloadSummaryStatistics:
    """
    BETA API - This API is still evolving.

    API to download job attachments based on given list of manifests.
    If path mapping rules file is given, map to corresponding destinations.

    Args:
        manifests (List[str]): File Path to the manifest file for upload.
        s3_root_uri (str): S3 root uri including bucket name and root prefix.
        boto3_session (boto3.Session): Boto3 session for interacting with customer s3.
        path_mapping_rules (Optional[str], optional): Optional file path to a JSON file contains list of path mapping. Defaults to None.
        print_function_callback (Callable[[str], None], optional): Callback function to provide visibility. Defaults to lambda msg: None.

    Raises:
        NonValidInputError: raise when any of the input is not valid.
    """

    file_name_manifest_dict: Dict[str, BaseAssetManifest] = _read_manifests(
        manifest_paths=manifests
    )

    path_mapping_rule_list: List[PathMappingRule] = _process_path_mapping(
        path_mapping_rules=path_mapping_rules
    )

    merged_manifests_by_root: Dict[str, BaseAssetManifest] = dict()
    for file_name, manifest in file_name_manifest_dict.items():
        # File name is supposed to be prefixed by a hash of source path in path mapping, use that to determine destination
        # If it doesn't appear in path mapping or mapping doesn't exist, download to current directory instead
        destination = next(
            (
                rule.destination_path
                for rule in (path_mapping_rule_list or [])
                if rule.get_hashed_source_path(manifest.get_default_hash_alg()) in file_name
            ),
            # Write to current directory partitioned by manifest name when no path mapping defined
            f"{os.getcwd()}/{file_name}",
        )
        # Assuming the manifest is already aggregated and correspond to a single destination
        if merged_manifests_by_root.get(destination):
            raise NonValidInputError(
                f"{destination} is already in use, one destination path maps to one manifest file only."
            )

        merged_manifests_by_root[destination] = manifest

    # Given manifests and S3 bucket + root, downloads all files from a CAS in each manifest.
    s3_settings: JobAttachmentS3Settings = JobAttachmentS3Settings.from_s3_root_uri(s3_root_uri)
    return download_files_from_manifests(
        s3_bucket=s3_settings.s3BucketName,
        manifests_by_root=merged_manifests_by_root,
        cas_prefix=s3_settings.full_cas_prefix(),
        session=boto3_session,
        conflict_resolution=conflict_resolution,
    )


def _attachment_upload(
    manifests: List[str],
    s3_root_uri: str,
    boto3_session: boto3.Session,
    root_dirs: List[str] = [],
    path_mapping_rules: Optional[str] = None,
    manifest_path_mapping: Optional[Dict[str, str]] = None,
    upload_manifest_path: Optional[str] = None,
    print_function_callback: Callable[[Any], None] = lambda msg: None,
) -> List[UploadManifestInfo]:
    """
    BETA API - This API is still evolving.

    API to upload job attachments based on given list of manifests and corresponding file directories.
    If path mapping rules file is given, map to corresponding destinations.

    Args:
        manifests (List[str]): File Path to the manifest file for upload.
        s3_root_uri (str): S3 root uri including bucket name and root prefix.
        boto3_session (boto3.Session): Boto3 session for interacting with customer s3.
        root_dirs (List[str]): List of root directories holding attachments. Defaults to empty.
        path_mapping_rules (Optional[str], optional): Optional file path to a JSON file contains list of path mapping. Defaults to None.
        upload_manifest_path (Optional[str], optional): Optional path prefix for uploading given manifests. Defaults to None.
        print_function_callback (Callable[[str], None], optional): Callback function to provide visibility. Defaults to lambda msg: None.

    Returns:
        List[UploadManifestInfo]: A list of UploadManifestInfo objects corresponding to the input manifests
        containing manifest path, hash information, and source path

    Raises:
        NonValidInputError: raise when any of the input is not valid.
    """

    file_name_manifest_dict: Dict[str, BaseAssetManifest] = _read_manifests(
        manifest_paths=manifests
    )

    if bool(path_mapping_rules) == bool(root_dirs):
        raise NonValidInputError("One of path mapping rule and root dir must exist, and not both.")

    path_mapping_rule_list: List[PathMappingRule] = _process_path_mapping(
        path_mapping_rules=path_mapping_rules, root_dirs=root_dirs
    )

    # Initialize an empty list to store manifest information
    manifest_info_list = []

    s3_settings: JobAttachmentS3Settings = JobAttachmentS3Settings.from_s3_root_uri(s3_root_uri)
    asset_uploader: S3AssetUploader = S3AssetUploader(session=boto3_session)

    # Iterate over original manifests in the order they were provided
    for manifest_path in manifests:
        file_name = os.path.basename(manifest_path)
        manifest: BaseAssetManifest = file_name_manifest_dict[file_name]

        # File name is supposed to be prefixed by a hash of source path in path mapping or provided root dirs
        rule: Optional[PathMappingRule] = next(
            # search in path mapping to determine source and destination
            (
                rule
                for rule in path_mapping_rule_list
                if rule.get_hashed_source_path(manifest.get_default_hash_alg()) in file_name
            ),
            None,
        )
        if not rule:
            raise NonValidInputError(
                f"No valid root defined for given manifest {file_name}, please check input root dirs and path mapping rule."
            )

        metadata = {"Metadata": {"asset-root": json.dumps(rule.source_path, ensure_ascii=True)}}
        # S3 metadata must be ASCII, so use either 'asset-root' or 'asset-root-json' depending
        # on whether the value is ASCII.
        try:
            # Add the 'asset-root' metadata if the path is ASCII
            rule.source_path.encode(encoding="ascii")
            metadata["Metadata"]["asset-root"] = rule.source_path
        except UnicodeEncodeError:
            # Add the 'asset-root-json' metadata encoded to ASCII as a JSON string
            metadata["Metadata"]["asset-root-json"] = json.dumps(
                rule.source_path, ensure_ascii=True
            )
        if rule.source_path_format:
            metadata["Metadata"]["file-system-location-name"] = rule.source_path_format

        # Uploads all files to a CAS in the manifest, optionally upload manifest file
        key, data = asset_uploader.upload_assets(
            job_attachment_settings=s3_settings,
            manifest=manifest,
            partial_manifest_prefix=upload_manifest_path,
            manifest_file_name=file_name,
            manifest_metadata=metadata,
            source_root=Path(rule.source_path),
            asset_root=Path(rule.destination_path),
            s3_check_cache_dir=config_file.get_cache_directory(),
        )
        print_function_callback(
            f"Uploaded assets from {rule.destination_path}, to {s3_settings.to_s3_root_uri()}/Manifests/{key}, hashed data {data}"
        )

        manifest_info_list.append(
            UploadManifestInfo(
                output_manifest_path=key,
                output_manifest_hash=data,
                source_path=rule.source_path,
            )
        )

    return manifest_info_list


def _process_path_mapping(
    path_mapping_rules: Optional[str] = None, root_dirs: List[str] = []
) -> List[PathMappingRule]:
    """
    Process list of path mapping rules from the input path mapping file or root directories.

    Args:
        path_mapping_rules (Optional[str], optional): File path to path mapping rules. Defaults to None.
        root_dirs (List[str], optional): List of root directories path. Defaults to [].

    Raises:
        NonValidInputError: Raise if any of the path mapping rule file or root dirs are not valid.

    Returns:
        List[PathMappingRule]: List of processed PathMappingRule
    """

    path_mapping_rule_list: List[PathMappingRule] = list()

    if path_mapping_rules:
        if not os.path.isfile(path_mapping_rules):
            raise NonValidInputError(
                f"Specified path mapping file {path_mapping_rules} is not valid."
            )
        with open(path_mapping_rules, encoding="utf8") as f:
            data = json.load(f)
            if "path_mapping_rules" in data:
                data = data["path_mapping_rules"]

            assert isinstance(data, list), "Path mapping rules have to be a list of dict."
            path_mapping_rule_list.extend([PathMappingRule(**mapping) for mapping in data])

    if nonvalid_dirs := [root for root in root_dirs if not os.path.isdir(root)]:
        raise NonValidInputError(f"Specified root dir {nonvalid_dirs} are not valid.")

    path_mapping_rule_list.extend(
        PathMappingRule(source_path_format="", source_path=root, destination_path=root)
        for root in root_dirs
    )

    return path_mapping_rule_list



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Tool registry and mapping definitions for MCP tools in Deadline Cloud."""

from typing import Any, Callable, List, Optional, TypedDict, Dict

from ..client import api
from .tools import job


class ToolDefinition(TypedDict):
    """Definition of a single MCP tool including its function and parameters."""

    func: Callable[..., Any]
    param_names: Optional[List[str]]


def get_tool_definition(tool_name: str) -> ToolDefinition:
    """Get the definition for a specific tool."""
    if tool_name not in TOOL_REGISTRY:
        raise ValueError(f"Tool '{tool_name}' not found in registry")
    return TOOL_REGISTRY[tool_name]


def get_all_tool_names() -> List[str]:
    """Get all registered tool names."""
    return list(TOOL_REGISTRY.keys())


TOOL_REGISTRY: Dict[str, ToolDefinition] = {
    "list_farms": {
        "func": api.list_farms,
        "param_names": ["nextToken", "principalId", "maxResults"],
    },
    "list_queues": {
        "func": api.list_queues,
        "param_names": ["farmId", "principalId", "status", "nextToken", "maxResults"],
    },
    "list_jobs": {
        "func": api.list_jobs,
        "param_names": ["farmId", "queueId", "principalId", "nextToken", "maxResults"],
    },
    "list_fleets": {
        "func": api.list_fleets,
        "param_names": [
            "farmId",
            "principalId",
            "displayName",
            "status",
            "nextToken",
            "maxResults",
        ],
    },
    "list_storage_profiles_for_queue": {
        "func": api.list_storage_profiles_for_queue,
        "param_names": ["farmId", "queueId", "nextToken", "maxResults"],
    },
    "check_authentication_status": {
        "func": api.check_authentication_status,
        "param_names": None,
    },
    "get_session_logs": {
        "func": api.get_session_logs,
        "param_names": [
            "farm_id",
            "queue_id",
            "session_id",
            "job_id",
            "limit",
            "start_time",
            "end_time",
            "next_token",
        ],
    },
    "submit_job": {
        "func": job.submit_job,
        "param_names": None,
    },
    "download_job_output": {
        "func": job.download_job_output,
        "param_names": None,
    },
    # Diagnostics - Primitive APIs
    "get_job": {
        "func": api.get_job,
        "param_names": ["farm_id", "queue_id", "job_id"],
    },
    "get_session": {
        "func": api.get_session,
        "param_names": ["farm_id", "queue_id", "job_id", "session_id"],
    },
    "list_sessions": {
        "func": api.list_sessions,
        "param_names": ["farm_id", "queue_id", "job_id", "max_results"],
    },
    "list_steps": {
        "func": api.list_steps,
        "param_names": ["farm_id", "queue_id", "job_id", "max_results"],
    },
    "list_tasks": {
        "func": api.list_tasks,
        "param_names": ["farm_id", "queue_id", "job_id", "step_id", "max_results"],
    },
    "search_jobs": {
        "func": api.search_jobs,
        "param_names": [
            "farm_id",
            "queue_ids",
            "task_run_status",
            "name_contains",
            "page_size",
            "item_offset",
        ],
    },
}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Private helper functions for job CLI commands.
"""

from configparser import ConfigParser
from datetime import datetime
from typing import Optional

import click
from botocore.exceptions import ClientError

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError
from .._common import _cli_object_repr, _suggest_resources_on_client_error


def _format_timestamp(dt: datetime) -> str:
    """Format a datetime to local time with timezone offset."""
    if dt is None:
        return ""
    local_dt = dt.astimezone()
    return local_dt.strftime("%Y-%m-%d %H:%M:%S %z")


def _truncate_middle(text: str, max_length: int) -> str:
    """Truncate text in the middle with '...' if too long."""
    if len(text) <= max_length:
        return text
    # Keep more at start than end for readability
    keep = max_length - 3  # account for "..."
    start = (keep * 2) // 3
    end = keep - start
    return text[:start] + "..." + text[-end:]


def _format_task_summary(task_counts: dict) -> str:
    """Format task status counts into a brief summary."""
    ready = task_counts.get("READY", 0)
    running = (
        task_counts.get("RUNNING", 0)
        + task_counts.get("STARTING", 0)
        + task_counts.get("ASSIGNED", 0)
        + task_counts.get("SCHEDULED", 0)
    )
    interrupting = task_counts.get("INTERRUPTING", 0)
    pending = task_counts.get("PENDING", 0)
    suspended = task_counts.get("SUSPENDED", 0)
    succeeded = task_counts.get("SUCCEEDED", 0)
    failed = task_counts.get("FAILED", 0)
    canceled = task_counts.get("CANCELED", 0)
    not_compatible = task_counts.get("NOT_COMPATIBLE", 0)

    parts = []
    if ready:
        parts.append(f"{ready} ready")
    if running:
        parts.append(f"{running} running")
    if interrupting:
        parts.append(f"{interrupting} interrupting")
    if pending:
        parts.append(f"{pending} pending")
    if suspended:
        parts.append(f"{suspended} suspended")
    if succeeded:
        parts.append(f"{succeeded} succeeded")
    if failed:
        parts.append(f"{failed} failed")
    if canceled:
        parts.append(f"{canceled} canceled")
    if not_compatible:
        parts.append(f"{not_compatible} not compatible")

    return ", ".join(parts) if parts else "no tasks"


def _resolve_job_search(config: Optional[ConfigParser], search_term: str) -> Optional[str]:
    """
    Search for jobs with a search term and resolve to a single job ID.

    Returns the job ID if exactly one job matches.
    If multiple jobs match, prints a summary table and returns None.
    If no jobs match, prints a message and returns None.
    """
    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)

    try:
        response = deadline.search_jobs(
            farmId=farm_id,
            queueIds=[queue_id],
            itemOffset=0,
            pageSize=5,
            filterExpressions={
                "filters": [
                    {"searchTermFilter": {"searchTerm": search_term, "matchType": "CONTAINS"}}
                ],
                "operator": "AND",
            },
            sortExpressions=[{"fieldSort": {"name": "CREATED_AT", "sortOrder": "DESCENDING"}}],
        )
    except ClientError as exc:
        raise DeadlineOperationError(f"Failed to search jobs:\n{exc}") from exc

    jobs = response.get("jobs", [])
    total = response.get("totalResults", 0)

    if not jobs:
        click.echo(f'No jobs found matching "{search_term}"')
        return None

    # Single result - return job ID for detailed lookup
    if total == 1:
        return jobs[0]["jobId"]

    # Multiple results - display summary
    click.echo(f'Found {total} job(s) matching "{search_term}", showing most recent {len(jobs)}:\n')

    for job in jobs:
        name = _truncate_middle(job.get("name", job.get("displayName", "")), 80)
        created = _format_timestamp(job.get("createdAt"))
        task_counts = job.get("taskRunStatusCounts", {})
        task_summary = _format_task_summary(task_counts)

        click.echo(f"  {name}")
        click.echo(f"    {job['jobId']}  {job['taskRunStatus']:<12}  {created}")
        click.echo(f"    Tasks: {task_summary}")
        click.echo()

    if total > len(jobs):
        click.echo(f"\n  ... and {total - len(jobs)} more")

    click.echo("\nTo get details, run: deadline job get --job-id <job-id>")
    return None


def _print_job_details(config: Optional[ConfigParser], job_id: str) -> None:
    """Fetch and print full job details."""
    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    try:
        response = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    except ClientError as exc:
        suggestion = _suggest_resources_on_client_error(
            exc, farm_id=farm_id, queue_id=queue_id, config=config
        )
        raise DeadlineOperationError(
            f"Failed to get Job from Deadline:\n{exc}{suggestion}"
        ) from exc
    response.pop("ResponseMetadata", None)
    click.echo(_cli_object_repr(response))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Contains the base asset manifest and entities that are part of the Asset Manifest"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Any, ClassVar

from .hash_algorithms import HashAlgorithm
from .versions import ManifestVersion


@dataclass
class BaseManifestPath(ABC):
    """
    Data class for paths in the Asset Manifest
    """

    path: str
    hash: str
    size: int
    mtime: int
    manifest_version: ClassVar[ManifestVersion]

    def __init__(self, *, path: str, hash: str, size: int, mtime: int) -> None:
        self.path = path
        self.hash = hash
        self.size = size
        self.mtime = mtime

    def __eq__(self, other: object) -> bool:
        """
        By default dataclasses still check ClassVars for equality.
        We only want to compare fields.
        :param other:
        :return: True if all fields are equal, False otherwise.
        """
        if not isinstance(other, BaseManifestPath):
            return NotImplemented
        return fields(self) == fields(other)


@dataclass
class BaseAssetManifest(ABC):
    """Base class for the Asset Manifest."""

    hashAlg: HashAlgorithm
    paths: list[BaseManifestPath]
    manifestVersion: ManifestVersion

    def __init__(
        self,
        *,
        paths: list[BaseManifestPath],
        hash_alg: HashAlgorithm,
    ):
        self.paths = paths
        self.hashAlg = hash_alg

    @classmethod
    @abstractmethod
    def get_default_hash_alg(cls) -> HashAlgorithm:  # pragma: no cover
        """Returns the default hashing algorithm for the Asset Manifest"""
        raise NotImplementedError(
            "Asset Manifest base class does not implement get_default_hash_alg"
        )

    @classmethod
    @abstractmethod
    def decode(cls, *, manifest_data: dict[str, Any]) -> BaseAssetManifest:  # pragma: no cover
        """Turn a dictionary for a manifest into an AssetManifest object"""
        raise NotImplementedError("Asset Manifest base class does not implement decode")

    @abstractmethod
    def encode(self) -> str:  # pragma: no cover
        """
        Recursively encode the Asset Manifest into a string according to
        whatever format the Asset Manifest was written for.
        """
        raise NotImplementedError("Asset Manifest base class does not implement encode")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = [
    "bundle_group",
    "config_group",
    "auth_group",
    "farm_group",
    "fleet_group",
    "handle_web_url_command",
    "job_group",
    "queue_group",
    "worker_group",
    "attachment_group",
    "manifest_group",
    "mcp_server_command",
]

from . import (
    bundle_group as bundle_group,
    config_group as config_group,
    auth_group as auth_group,
    farm_group as farm_group,
    fleet_group as fleet_group,
    handle_web_url_command as handle_web_url_command,
    job_group as job_group,
    queue_group as queue_group,
    worker_group as worker_group,
    attachment_group as attachment_group,
    manifest_group as manifest_group,
    mcp_server_command as mcp_server_command,
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os

from contextlib import ExitStack
from typing import List, Dict

from ...client.exceptions import NonValidInputError
from ..asset_manifests.base_manifest import BaseAssetManifest
from ..asset_manifests.decode import decode_manifest


def _read_manifests(manifest_paths: List[str]) -> Dict[str, BaseAssetManifest]:
    """
    Read in manfiests from the given file path list, and produce file name to manifest mapping.

    Args:
        manifest_paths (List[str]): List of file paths to manifest file.

    Raises:
        NonValidInputError: Raise when any of the file is not valid.

    Returns:
        Dict[str, BaseAssetManifest]: File name to encoded manifest mapping
    """

    if nonvalid_files := [manifest for manifest in manifest_paths if not os.path.isfile(manifest)]:
        raise NonValidInputError(f"Specified manifests {nonvalid_files} are not valid.")

    with ExitStack() as stack:
        file_name_manifest_dict: Dict[str, BaseAssetManifest] = {
            os.path.basename(file_path): decode_manifest(
                stack.enter_context(open(file_path)).read()
            )
            for file_path in manifest_paths
        }

    return file_name_manifest_dict



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .. import api
from ..config import config_file
from ...job_attachments.models import AssetRootGroup, AssetRootManifest
from ...job_attachments.upload import S3AssetManager, SummaryStatistics
from ...job_attachments.progress_tracker import ProgressReportMetadata, ProgressStatus


import textwrap
from configparser import ConfigParser
from typing import Callable, List, Optional, Tuple


def _hash_attachments(
    asset_manager: S3AssetManager,
    asset_groups: List[AssetRootGroup],
    total_input_files: int,
    total_input_bytes: int,
    print_function_callback: Callable = lambda msg: None,
    hashing_progress_callback: Optional[Callable] = None,
    config: Optional[ConfigParser] = None,
) -> Tuple[SummaryStatistics, List[AssetRootManifest]]:
    """
    Starts the job attachments hashing and handles the progress reporting
    callback. Returns a list of the asset manifests of the hashed files.
    """

    def _default_update_hash_progress(hashing_metadata: ProgressReportMetadata) -> bool:
        return True

    if not hashing_progress_callback:
        hashing_progress_callback = _default_update_hash_progress

    hashing_summary, manifests = asset_manager.hash_assets_and_create_manifest(
        asset_groups=asset_groups,
        total_input_files=total_input_files,
        total_input_bytes=total_input_bytes,
        hash_cache_dir=config_file.get_cache_directory(),
        on_preparing_to_submit=hashing_progress_callback,
    )
    api.get_deadline_cloud_library_telemetry_client(config=config).record_hashing_summary(
        hashing_summary
    )
    if hashing_summary.total_files > 0:
        print_function_callback("Hashing Summary:")
        print_function_callback(textwrap.indent(str(hashing_summary), "    "))
    else:
        # Ensure to call the callback once if no files were processed
        hashing_progress_callback(
            ProgressReportMetadata(
                status=ProgressStatus.PREPARING_IN_PROGRESS,
                progress=100,
                transferRate=0,
                progressMessage="No files to hash",
                processedFiles=0,
            )
        )

    return hashing_summary, manifests



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Module for accessing the local 'last seen on S3' cache.
"""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from .cache_db import CacheDB


logger = logging.getLogger("Deadline")


@dataclass
class S3CheckCacheEntry:
    """Represents an entry in the local s3 check cache database"""

    s3_key: str
    last_seen_time: str

    def to_dict(self) -> Dict[str, Any]:
        return {
            "s3_key": self.s3_key,
            "last_seen_time": self.last_seen_time,
        }


class S3CheckCache(CacheDB):
    """
    Maintains a cache of 'last seen on S3' entries in a local database, which
    specifies which full S3 object keys exist in the content-addressed storage
    in the Job Attachments S3 bucket.

    This class is intended to always be used with a context manager to properly
    close the connection to the hash cache database.

    This class also automatically locks when doing writes, so it can be called
    by multiple threads.
    """

    CACHE_NAME = "s3_check_cache"
    CACHE_DB_VERSION = 1
    ENTRY_EXPIRY_DAYS = 30

    def __init__(self, cache_dir: Optional[str] = None) -> None:
        table_name: str = f"s3checkV{self.CACHE_DB_VERSION}"
        create_query: str = f"CREATE TABLE s3checkV{self.CACHE_DB_VERSION}(s3_key text primary key, last_seen_time timestamp)"
        super().__init__(
            cache_name=self.CACHE_NAME,
            table_name=table_name,
            create_query=create_query,
            cache_dir=cache_dir,
        )

    def get_connection_entry(self, s3_key: str, connection) -> Optional[S3CheckCacheEntry]:
        """
        Returns an entry from the hash cache, if it exists.  This is the "lockless" (Doesn't take
        the main db_lock protecting db_connection) version of get_entry which expects a connection
        parameter for the connection which will be used to read from the DB - this can generally
        be the thread local connection returned by get_local_connection()
        """

        entry_vals = connection.execute(
            f"SELECT * FROM {self.table_name} WHERE s3_key=?",
            [s3_key],
        ).fetchone()
        if entry_vals:
            entry = S3CheckCacheEntry(
                s3_key=entry_vals[0],
                last_seen_time=str(entry_vals[1]),
            )
            try:
                last_seen = datetime.fromtimestamp(float(entry.last_seen_time))
                if (datetime.now() - last_seen).days < self.ENTRY_EXPIRY_DAYS:
                    return entry
            except ValueError:
                logger.warning(f"Timestamp for S3 key {s3_key} is not valid. Ignoring.")

        return None

    def get_entry(self, s3_key: str) -> Optional[S3CheckCacheEntry]:
        """
        Checks if an entry exists in the cache, and returns it if it hasn't expired.
        """
        if not self.enabled:
            return None

        with self.db_lock, self.db_connection:
            return self.get_connection_entry(s3_key, self.db_connection)

    def put_entry(self, entry: S3CheckCacheEntry) -> None:
        """Inserts or replaces an entry into the cache database."""
        if self.enabled:
            with self.db_lock, self.db_connection:
                self.db_connection.execute(
                    f"INSERT OR REPLACE INTO {self.table_name} VALUES(:s3_key, :last_seen_time)",
                    entry.to_dict(),
                )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

__all__ = ["DecimalMode", "FloatDragSpinBox", "IntDragSpinBox"]

import enum
from math import ceil, floor, log10

from qtpy.QtCore import QEvent, QObject, Qt
from qtpy.QtGui import QCursor, QKeyEvent, QMouseEvent
from qtpy.QtWidgets import QAbstractSpinBox, QDoubleSpinBox, QSpinBox, QWidget


class DecimalMode(enum.Enum):
    """
    Defines the valid Decimal Modes that can be used with a FloatDragSpinBox and
    IntDragSpinBox.
    """

    FIXED_DECIMAL = enum.auto()
    ADAPTIVE_DECIMAL = enum.auto()


class FloatDragSpinBox(QDoubleSpinBox):
    """
    A improved QDoubleSpinBox that adds the ability to click and drag the
    SpinBox arrow buttons to increment/decrement the value. This Widget also
    adds the ability to adaptively set the number of decimal places based on the
    size of the value.
    """

    MAX_ADAPTIVE_DECIMALS = 5
    MAX_FLOAT_VALUE = 1e308
    MIN_FLOAT_VALUE = -1e308

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setMouseTracking(False)
        self.setCursor(Qt.SizeVerCursor)
        self.setMaximum(self.MAX_FLOAT_VALUE)
        self.setMinimum(self.MIN_FLOAT_VALUE)
        self.setDragMultiplier(0.1)
        self.setDecimalMode(DecimalMode.ADAPTIVE_DECIMAL)
        self.setStepType(QDoubleSpinBox.AdaptiveDecimalStepType)
        self.lineEdit().installEventFilter(self)

        self.editingFinished.connect(self._stop_editing)
        self.valueChanged.connect(self._set_adaptive_decimals)

        self._cursor_start_pos = self.mapToGlobal(self.pos())
        self._editing = False
        self._dragging = False

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Hides the cursor on click and records the current cursor position to be
        used if the user then decides to start dragging the mouse.
        """
        super().mousePressEvent(event)

        self.setCursor(Qt.BlankCursor)
        self._cursor_start_pos = event.globalPos()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """
        Sets the Widget's value based on the position difference that the cursor
        was moved while holding the mouse button down.
        """
        super().mouseMoveEvent(event)

        self._dragging = True
        cursor_current_pos = event.globalPos()
        cursor_offset = self._cursor_start_pos.y() - cursor_current_pos.y()
        new_value = self.value() + (cursor_offset * self.dragMultiplier())
        new_value = min(new_value, self.maximum())
        new_value = max(new_value, self.minimum())

        self.setValue(new_value)
        QCursor.setPos(self._cursor_start_pos)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """
        Makes the cursor visible again once the mouse click is released and sets
        that the user is no longer dragging the mouse.
        """
        super().mouseReleaseEvent(event)

        self.setCursor(Qt.SizeVerCursor)
        self._dragging = False

    def stepEnabled(self) -> QAbstractSpinBox.StepEnabled:  # type: ignore[name-defined]
        """
        Disables the arrow buttons if the user is using the hold and drag
        functionaltiy to change the value.
        """
        if self._dragging:
            return QAbstractSpinBox.StepNone
        return super().stepEnabled()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Captures any key press events to set that the user is likely manually
        editing the QLineEdit that is part of the Widget.
        """
        self._start_editing()
        super().keyPressEvent(event)

    def eventFilter(self, object: QObject, event: QEvent) -> bool:
        """
        Captures when the user clicks on the QLineEdit that is part of the
        Widget and tries to set that the user is currently editing manually.
        """
        if object == self.lineEdit() and event.type() == QEvent.MouseButtonPress:
            self._start_editing()
        return False

    def decimalMode(self) -> DecimalMode:
        return self._decimal_mode

    def setDecimalMode(self, mode: DecimalMode) -> None:
        self._decimal_mode = mode

    def dragMultiplier(self) -> float:
        return self._drag_multiplier

    def setDragMultiplier(self, multiplier: float) -> None:
        self._drag_multiplier = max(0.0, multiplier)

    def _set_adaptive_decimals(self, value: float) -> None:
        """
        Adaptively sets the number of decimal places to use based on the current
        value when ADAPTIVE_DECIMAL mode is being used.
        """
        if self.decimalMode() == DecimalMode.ADAPTIVE_DECIMAL:
            if value < -1 or value > 1:  # Don't take log10 of (-1,1)
                self.setDecimals(max(0, self.MAX_ADAPTIVE_DECIMALS - floor(abs(log10(abs(value))))))
            else:
                self.setDecimals(self.MAX_ADAPTIVE_DECIMALS)

    def _start_editing(self) -> None:
        """
        Sets the number of decimal places to the max if using ADAPTIVE_DECIMAL
        mode. This is done while the user is editing in the QLineEdit so they
        aren't prevented from entering a small value that would have more
        decimal places than the current adaptive decimal places.
        """
        if not self._editing:
            if self.decimalMode() == DecimalMode.ADAPTIVE_DECIMAL:
                self.setDecimals(self.MAX_ADAPTIVE_DECIMALS)
            self._editing = True

    def _stop_editing(self) -> None:
        """
        Tries to set the number of decimal places back to being adaptively
        calculated once the user is finished editing. If ADAPTIVE_DECIMAL mode
        is not being used then nothing will be changed.
        """
        self._set_adaptive_decimals(self.value())
        self._editing = False


class IntDragSpinBox(QSpinBox):
    """
    A improved QSpinBox that adds the ability to click and drag the
    SpinBox arrow buttons to increment/decrement the value.
    """

    MAX_INT_VALUE = (2**31) - 1
    MIN_INT_VALUE = -(2**31) + 1

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.setKeyboardTracking(False)
        self.setMouseTracking(False)
        self.setCursor(Qt.SizeVerCursor)
        self.setMaximum(self.MAX_INT_VALUE)
        self.setMinimum(self.MIN_INT_VALUE)
        self.setDragMultiplier(0.1)
        self.setStepType(QSpinBox.AdaptiveDecimalStepType)

        self._cursor_start_pos = self.mapToGlobal(self.pos())
        self._is_dragging = False

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Hides the cursor on click and records the current cursor position to be
        used if the user then decides to start dragging the mouse.
        """
        super().mousePressEvent(event)
        self.setCursor(Qt.BlankCursor)
        self._cursor_start_pos = event.globalPos()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """
        Sets the Widget's value based on the position difference that the cursor
        was moved while holding the mouse button down.
        """
        super().mouseMoveEvent(event)

        self._is_dragging = True

        cursor_current_pos = event.globalPos()
        cursor_offset = self._cursor_start_pos.y() - cursor_current_pos.y()
        new_value = self.value() + (cursor_offset * self.dragMultiplier())
        # Always increment / decrement by at least 1 regardless of the offset
        new_value = ceil(new_value) if cursor_offset > 0 else floor(new_value)
        new_value = min(new_value, self.maximum())
        new_value = max(new_value, self.minimum())

        self.setValue(new_value)
        QCursor.setPos(self._cursor_start_pos)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """
        Makes the cursor visible again once the mouse click is released and sets
        that the user is no longer dragging the mouse.
        """
        super().mouseReleaseEvent(event)
        self.setCursor(Qt.SizeVerCursor)
        self._is_dragging = False

    def stepEnabled(self) -> QAbstractSpinBox.StepEnabled:  # type: ignore[name-defined]
        """
        Disables the arrow buttons if the user is using the hold and drag
        functionaltiy to change the value.
        """
        if self._is_dragging:
            return QAbstractSpinBox.StepNone
        return super().stepEnabled()

    def dragMultiplier(self) -> float:
        return self._drag_multiplier

    def setDragMultiplier(self, multiplier: float) -> None:
        self._drag_multiplier = max(0.0, multiplier)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

__all__ = ["get_storage_profile_for_queue"]

from configparser import ConfigParser
from typing import Optional
from botocore.client import BaseClient  # type: ignore[import]

from ._session import get_boto3_client
from .. import api
from ...job_attachments.models import (
    FileSystemLocation,
    FileSystemLocationType,
    StorageProfile,
    StorageProfileOperatingSystemFamily,
)


@api.record_function_latency_telemetry_event()
def get_storage_profile_for_queue(
    farm_id: str,
    queue_id: str,
    storage_profile_id: str,
    deadline: Optional[BaseClient] = None,
    config: Optional[ConfigParser] = None,
) -> StorageProfile:
    if deadline is None:
        deadline = get_boto3_client("deadline", config=config)

    storage_profile_response = deadline.get_storage_profile_for_queue(
        farmId=farm_id, queueId=queue_id, storageProfileId=storage_profile_id
    )
    return StorageProfile(
        storageProfileId=storage_profile_response["storageProfileId"],
        displayName=storage_profile_response["displayName"],
        osFamily=StorageProfileOperatingSystemFamily(storage_profile_response["osFamily"]),
        fileSystemLocations=[
            FileSystemLocation(
                name=file_system_location["name"],
                path=file_system_location["path"],
                type=FileSystemLocationType(file_system_location["type"]),
            )
            for file_system_location in storage_profile_response.get("fileSystemLocations", [])
        ],
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations
import os
import re
from typing import Any, Optional, Iterable, Union
from collections.abc import Collection
from pathlib import PurePath, PurePosixPath, PureWindowsPath
import ntpath
import posixpath

from ..job_attachments.models import PathFormat


def human_readable_file_size(size_in_bytes: int) -> str:
    """
    Convert a size in bytes to something human readable. For example 1000 bytes will be converted
    to 1 KB. Sizes close enough to a postfix threshold will be rounded up to the next threshold.
    For example 999999 bytes would be output as 1.0 MB and NOT 999.99 KB (or as a consequence of
    Python's round function 1000.0 KB).

    This function is for display purposes only.
    """
    converted_size: Union[int, float] = size_in_bytes
    rounded: Union[int, float]
    postfixes = ["B", "KB", "MB", "GB", "TB", "PB"]

    for postfix in postfixes:
        rounded = round(converted_size, ndigits=2)

        if rounded < 1000:
            return f"{rounded} {postfix}"

        converted_size /= 1000

    # If we go higher than the provided postfix,
    # then return as a large amount of the highest postfix we've specified.
    return f"{rounded} {postfixes[-1]}"


_NUMBERED_PATH_REGEX = re.compile(r"^(.*\D|)(\d+)(\.[^/\\]+)?$")


class _NumberedPath:
    """
    Representation of a file system path that may be numbered like frame_001.png.

    Some example properties for paths:
      frame_001.png: grouping='frame_#.png', padding_min=3, padding_max=3, number=1
      sequence_v907: grouping='sequence_v#', padding_min=1, padding_max=3, number=907
    """

    path: str
    """The path that may or may not be numbered"""
    parts: Optional[list[str]] = None
    """The path split into parts"""
    grouping: str
    """The path with the number as '#' for grouping purposes"""
    padding_min: int
    """The minimum padding or -1. e.g. '%04d' for padding of 4, that produces the number in the path"""
    padding_max: int
    """The maximum padding or -1. e.g. '%04d' for padding of 4, that produces the number in the path"""
    number: Optional[int] = None
    """The number in the path, or None if the path is not numbered"""

    def __init__(self, path: str):
        m = _NUMBERED_PATH_REGEX.match(path)
        if m:
            self.path = path
            self.parts = [m.group(1), m.group(2), m.group(3) or ""]
            self.grouping = f"{self.parts[0]}#.{self.parts[2]}"
            number = self.parts[1]
            if number[0] == "0":
                self.padding_min = len(number)
            else:
                self.padding_min = 1
            self.padding_max = len(number)
            self.number = int(number)
        else:
            self.path = self.grouping = path
            self.padding_min = self.padding_max = -1

    def __repr__(self):
        return f"_NumberedPath({self.path!r})"


def _divide_numbered_path_group(group: list[_NumberedPath]) -> dict[str, set[int]]:
    """Given a list of numbered paths that all have the same grouping string, check
    for padding consistency and split into multiple groups if necessary. Convert
    into a dictionary from a printf pattern to the set of numbers for it. Groups
    of size 2 are divided into individual paths.

    For example, the paths frame_001.png and frame_0002.png cannot be together,
    because they require different padding values."""

    result: dict[str, set[int]] = {}

    while len(group) > 0:
        # Treat groups of size 1 or 2 as individual paths
        if len(group) <= 2:
            for numbered_path in group:
                result[numbered_path.path] = set()
            break

        # The largest minimum padding is likely the right padding for the group
        padding = max(numbered_path.padding_min for numbered_path in group)
        pattern = f"%0{padding}d" if padding > 1 else "%d"
        consistent_group = [
            numbered_path for numbered_path in group if numbered_path.padding_max >= padding
        ]
        pattern_path = f"{consistent_group[0].parts[0]}{pattern}{consistent_group[0].parts[2]}"  # type: ignore
        result[pattern_path] = {numbered_path.number for numbered_path in consistent_group}  # type: ignore
        # Process the remaining paths separately
        group = [path for path in group if path.padding_max < padding]

    return result


class PathSummary:
    """
    Represents a summary of a path, including the path itself, the number of files,
    and the total size. The summary represents a sequence of files when the index_set
    value is non-empty. If the summary is a nested accumulation of paths, child
    paths are in the dictionary 'children'.

    If a path represents a directory, it ends with a directory separator.
    """

    path: str
    """Either the path, or a printf pattern if index_set is non-empty"""
    index_set: set[int]
    """The set of indexes if the path is a printf pattern or an empty set otherwise"""
    file_count: int
    """The number of files"""
    total_size: Optional[int]
    """The total size of all files, if sizes are provided"""
    children: Optional[dict[str, "PathSummary"]]
    """The children of this path, if the summary is nested"""

    _os_path: Any
    """Either ntpath or posixpath depending on the path_format from construction."""

    def __init__(
        self,
        path: str,
        *,
        path_format: Optional[PathFormat] = None,
        index_set: Optional[set[int]] = None,
        file_count: Optional[int] = None,
        total_size: Optional[int] = None,
        children: Optional[dict[str, "PathSummary"]] = None,
    ):
        if path_format is None:
            self._os_path = os.path
        elif path_format == PathFormat.WINDOWS:
            self._os_path = ntpath
        else:
            self._os_path = posixpath

        self.path = path
        self.index_set = index_set or set()
        if index_set:
            self.file_count = len(index_set)
        elif file_count is None:
            self.file_count = 0 if self.is_dir() else 1
        else:
            self.file_count = file_count
        self.total_size = total_size
        self.children = children

    def is_dir(self) -> bool:
        """Returns True if the path is a directory (indicated by a trailing '/')"""
        # On Windows, both '/' and '\\' are directory separators, so check both sep and altsep
        return self.path.endswith(self._os_path.sep) or (
            self._os_path.altsep and self.path.endswith(self._os_path.altsep)
        )  # type: ignore

    def summary(self, *, include_totals=True, relative_to: Optional[Union[PurePath, str]] = None):
        """Returns the path summary, including file count and size totals by default."""
        relpath = self.path
        if relative_to is not None:
            relpath = self._os_path.relpath(self.path, relative_to)
            # Ensure a trailing separator for directories
            if self.is_dir():
                relpath = self._os_path.join(relpath, "")

        if include_totals:
            return f"{relpath} ({self.summary_totals()})"
        elif self.index_set:
            return f"{relpath} (sequence indexes {_int_set_to_range_expr(self.index_set)})"
        else:
            return relpath

    def summary_totals(self) -> str:
        """Returns the totals of the summary, like 'sequence indexes 1-3, 3 files, 30 MB'
        if the path represents a sequence or '1 file' if the path represents a file
        and there is no size information available."""
        if self.index_set:
            seq_summary = f", sequence {_int_set_to_range_expr(self.index_set)}"
        else:
            seq_summary = ""
        if self.total_size is not None:
            size_summary = f", {human_readable_file_size(self.total_size)}"
        else:
            size_summary = ""
        plural = "s" if self.file_count != 1 else ""
        return f"{self.file_count} file{plural}{size_summary}{seq_summary}"

    def __str__(self):
        return self.summary()

    def __repr__(self):
        parts = ["PathSummary(", repr(self.path)]
        if self.is_dir():
            if self.file_count != 0:
                parts.append(f", file_count={self.file_count!r}")
        else:
            if self.index_set:
                parts.append(f", index_set={{{', '.join(str(v) for v in sorted(self.index_set))}}}")
        if self.total_size is not None:
            parts.append(f", total_size={self.total_size!r}")
        if self.children is not None:
            parts.append(f", children={self.children!r}")
        parts.append(")")
        return "".join(parts)

    def __eq__(self, value):
        if isinstance(value, PathSummary):
            return (
                self.path == value.path
                and self.index_set == value.index_set
                and self.file_count == value.file_count
                and self.total_size == value.total_size
                and self.children == value.children
            )
        else:
            return False


def _int_set_to_range_expr(int_set: set[int]) -> str:
    """
    Converts a set of integers into a range expression.
    For example, {1,2,3,4,5,7,8,9,10} -> "1-5,7-10"
    """
    int_list = sorted(set(int_set))
    range_expr_components = []
    last_interval_start = last_interval_end = int_list[0]

    def add_interval(start: int, end: int):
        if start == last_interval_end:
            range_expr_components.append(str(start))
        else:
            range_expr_components.append(f"{start}-{end}")

    for value in int_list[1:]:
        if value == last_interval_end + 1:
            last_interval_end = value
        else:
            add_interval(last_interval_start, last_interval_end)
            last_interval_start = last_interval_end = value
    add_interval(last_interval_start, last_interval_end)
    return ",".join(range_expr_components)


def summarize_paths_by_sequence(
    path_list: Collection[Union[PurePath, str]],
    *,
    path_format: Optional[PathFormat] = None,
    total_size_by_path: Optional[dict[str, int]] = None,
) -> list[PathSummary]:
    """
    Identifies numbered sequences of files/directories within a list of paths.
    Returns a sorted list of PathSummary objects. If total_size_by_path is provided, it
    must provide a total size for every path in path_list.

    >> group_sequence_paths(["frame_1.png", "frame_3.png", "frame_20.png", "readme.txt"])
    {PathSummary("frame_%d.png", index_set={1, 3, 20}), PathSummary("readme.txt")}

    >> group_sequence_paths(["frame_01.png", "frame_1.png", "frame_30.png", "frame_09.png"])
    {PathSummary("frame_%02d.png", index_set={1, 9, 20}), PathSummary("frame_1.png")}
    """
    if len(path_list) == 0:
        return []

    if path_format is None:
        path_format = PathFormat.get_host_path_format()

    # Convert all the paths into strings, and deduplicate by converting into a set
    path_list_as_str: set[str] = {
        str(path) if isinstance(path, PurePath) else path for path in path_list
    }
    # On Windows, convert all "/" separators to "\\"
    if path_format == PathFormat.WINDOWS:
        path_list_as_str = {path.replace("/", "\\") for path in path_list_as_str}
        if total_size_by_path:
            total_size_by_path = {
                path.replace("/", "\\"): size for path, size in total_size_by_path.items()
            }

    # Group according to the _NumberedPath.grouping property
    raw_grouped_paths: dict[str, list[_NumberedPath]] = {}
    for path in path_list_as_str:
        numbered_path = _NumberedPath(path)
        raw_grouped_paths.setdefault(numbered_path.grouping, []).append(numbered_path)

    # Divide any groups with inconsistent padding into smaller consistent groups,
    # and merge into a dictionary {printf_pattern: set of indexes}.
    grouped_paths: dict[str, set[int]] = {}
    for raw_group in raw_grouped_paths.values():
        grouped_paths.update(_divide_numbered_path_group(raw_group))

    # Sort the result by the printf pattern and convert to PathSummary objects
    result = [
        PathSummary(path, path_format=path_format, index_set=index_set)
        for path, index_set in sorted(grouped_paths.items(), key=lambda x: x[0])
    ]

    # If sizes are provided, populate them in the path summary objects
    if total_size_by_path:
        for path_summary in result:
            if path_summary.index_set:
                path_summary.total_size = sum(
                    total_size_by_path[path_summary.path % i] for i in path_summary.index_set
                )
            else:
                path_summary.total_size = total_size_by_path[path_summary.path]

    return result


def _collapse_each_path_summary(path_summary_list: Iterable[PathSummary]) -> list[PathSummary]:
    """
    Collapses each path summary in the list while it has a single child.
    """
    result = []
    for path_summary in path_summary_list:  # type: ignore
        while path_summary.children is not None and len(path_summary.children) == 1:
            path_summary = next(iter(path_summary.children.values()))
        result.append(path_summary)

    return result


def summarize_paths_by_nested_directory(
    path_list: Collection[Union[PurePath, str]],
    *,
    path_format: Optional[PathFormat] = None,
    total_size_by_path: Optional[dict[str, int]] = None,
) -> list[PathSummary]:
    """
    Summarizes the provided paths by sequence, and then nests them into
    common parent paths. The returned summaries do not contain a common parent,
    for example if they are different relative paths, or absolute paths for
    different drives on Windows

    By default, paths are for the current operating system. If path_format is provided, you can
    override that to PathFormat.WINDOWS or PathFormat.POSIX as necessary.
    """
    if len(path_list) == 0:
        return []

    if path_format is None:
        path_format = PathFormat.get_host_path_format()

    if path_format == PathFormat.WINDOWS:
        path_type: Any = PureWindowsPath
        os_path: Any = ntpath
    else:
        path_type = PurePosixPath
        os_path = posixpath

    # First summarize the paths by sequence
    summary_list = summarize_paths_by_sequence(
        path_list, total_size_by_path=total_size_by_path, path_format=path_format
    )

    # Put all the summaries into a temporary common root.
    nested_summary = PathSummary("ROOT/", path_format=path_format)
    for path_summary in summary_list:
        # Split the path into its components
        path_components = path_type(path_summary.path).parts
        # Start with the root component, and build up the nested structure
        current_level: PathSummary = nested_summary
        for i in range(len(path_components) - 1):
            component = path_components[i]
            # Add the child if it's not already there
            if current_level.children is None:
                current_level.children = {}
            if component not in current_level.children:
                current_level.children[component] = PathSummary(
                    os_path.join(*path_components[: i + 1], ""),
                    path_format=path_format,
                    total_size=0 if total_size_by_path else None,
                )
            # Descend into the new level
            current_level = current_level.children[component]
            # Accumulate the file counts and sizes
            current_level.file_count += path_summary.file_count
            if total_size_by_path:
                current_level.total_size += path_summary.total_size  # type: ignore
        # Add the path summary to the end
        if current_level.children is None:
            current_level.children = {}
        current_level.children[path_components[-1]] = path_summary

    # For each distinct root, collapse it while it contains a single child
    return _collapse_each_path_summary(nested_summary.children.values())  # type: ignore


def summarize_path_list(
    path_list: Collection[Union[PurePath, str]],
    *,
    path_format: Optional[PathFormat] = None,
    total_size_by_path: Optional[dict[str, int]] = None,
    max_entries=10,
    include_totals=True,
) -> str:
    """
    Creates a string summary of the files in the list provided,
    grouping numbered filenames by their sequence pattern, and nesting
    summaries of the directory into the specified maximum number of entries.

    By default, paths are for the current operating system. If path_format is provided, you can
    override that to PathFormat.WINDOWS or PathFormat.POSIX as necessary.

    If total_size_by_path is provided, it must provide a total size for every path in path_list.

        >>> print(summarize_path_list(["frame_1.png", "frame_3.png", "frame_20.png", "readme.txt"]))
        frame_%d.png (3 files, sequence 1,3,20)
        readme.txt (1 file)
    """
    if len(path_list) == 0:
        return ""

    if path_format is None:
        path_format = PathFormat.get_host_path_format()

    lines = []
    summary_list = summarize_paths_by_nested_directory(
        path_list, total_size_by_path=total_size_by_path, path_format=path_format
    )

    # If the summary list has one entry, and its path is a very shallow root like '/' or 'C:/',
    # then take all its children at the outer level. This makes the root paths longer so
    # the individually summarized paths will be shorter and easier to look through.
    if (
        len(summary_list) == 1
        and summary_list[0].children is not None
        and len(summary_list[0].children) <= max_entries / 2
    ):
        summary_list = _collapse_each_path_summary(summary_list[0].children.values())

    if total_size_by_path:
        # Sort the list so the largest size is first
        summary_list.sort(key=lambda v: (-v.total_size, v.path))  # type: ignore
    else:
        # Sort the list so the largest file count is first
        summary_list.sort(key=lambda v: (-v.file_count, v.path))  # type: ignore

    # Determine how many entries to show at the outer level and one level in,
    # with a total less than or equal to max_entries
    entry_counts = [
        0 if summary_path.children is None else min(len(summary_path.children), max_entries)
        for summary_path in summary_list[:max_entries]
    ]
    while len(entry_counts) + sum(entry_counts) > max_entries:
        max_entry_count = max(entry_counts)
        if max_entry_count > len(entry_counts):
            # If the largest entry count under a root path is more than the number of root paths,
            # then decrease that entry count.
            for i, entry_count in enumerate(reversed(entry_counts)):
                if entry_count == max_entry_count:
                    entry_counts[len(entry_counts) - i - 1] -= 1
                    break
        else:
            # Otherwise drop a root path from the summary
            entry_counts.pop()

    # If we're going to show "... and 1 more ..." after the items, might as well show
    # the last item instead
    if len(entry_counts) == len(summary_list) - 1 and not summary_list[-1].is_dir():
        entry_counts.append(0)

    for entry_count, summary_path in zip(entry_counts, summary_list):
        if summary_path.children is None:
            lines.append(f"{summary_path.summary(include_totals=include_totals)}\n")
        else:
            lines.append(f"{summary_path.summary(include_totals=include_totals)}:\n")
            children = list(summary_path.children.values())
            if total_size_by_path:
                # Sort the list so the largest size is first
                children.sort(key=lambda v: (-v.total_size, v.path))  # type: ignore
            else:
                # Sort the list so the largest file count is first
                children.sort(key=lambda v: (-v.file_count, v.path))  # type: ignore

            # If we're going to show "... and 1 more ..." after the items, might as well show
            # the last item instead
            if entry_count == len(children) - 1:
                entry_count += 1

            for child in children[:entry_count]:
                lines.append(
                    f"  {child.summary(include_totals=include_totals, relative_to=summary_path.path)}\n"
                )
            if len(summary_path.children) > entry_count:
                lines.append(f"  ... and {len(summary_path.children) - entry_count} more\n")

    if len(summary_list) > len(entry_counts):
        file_count = sum(v.file_count for v in summary_list[len(entry_counts) :])
        if total_size_by_path and include_totals:
            total_size = sum(v.total_size for v in summary_list[len(entry_counts) :])  # type: ignore
            lines.append(
                f"... and {len(summary_list) - len(entry_counts)} more ({file_count} files, {human_readable_file_size(total_size)})\n"
            )
        elif include_totals:
            lines.append(
                f"... and {len(summary_list) - len(entry_counts)} more ({file_count} files)\n"
            )
        else:
            lines.append(f"... and {len(summary_list) - len(entry_counts)} more\n")

    return "".join(lines)
